/*
	Copyright 2015 Barrett Technology <support@barrett.com>

	This file is part of libbarrett.

	This version of libbarrett is free software: you can redistribute it
	and/or modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	This version of libbarrett is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this version of libbarrett.  If not, see
	<http://www.gnu.org/licenses/>.

*/

/** Defines a boilerplate main() function that initializes a Proficio.
 *
 * This "standard" main() function simply initializes the Proficio in the normal way, by:
 *   - Waiting for the Proficio to be Shift-idled, if it's not already idled
 *   - Waking all Pucks found on the bus
 *   - Zero the Proficio, if it's not already zeroed
 *   - Waiting for the Proficio to be Shift-activated
 *   - Calling the proficio_main() function that the user is responsible for defining
 *   - Bypasses the waiting sequences that require the control pendant's input if a macro is defined
 *   .
 * The behavior is appropriate for many applications, but there is no issue with writing a custom main() function.
 *
 * The proficio_main() function must be able to accept a reference to a 3-DOF, 4-DOF and a 7-DOF barrett::systems::Wam. 
 *
 * @file standard_proficio_main.h
 * @date Dec 09, 2015
 * @author Hariharasudan Malaichamee
 */


#ifndef PROFICIO_STANDARD_MAIN_FUNCTION_H_
#define PROFICIO_STANDARD_MAIN_FUNCTION_H_


#include <barrett/exception.h>
#include <barrett/products/product_manager.h>
#include <barrett/systems/wam.h>
#include <barrett/systems.h>
#include <proficio/tools/bypass_safety_module.h>


enum config {LEFT, RIGHT};

#ifdef NO_CONTROL_PENDANT
	#define BARRETT_SMF_CONFIGURE_PM
	#define BARRETT_SMF_DONT_PROMPT_ON_ZEROING
	#define BARRETT_SMF_DONT_WAIT_FOR_SHIFT_ACTIVATE
#endif

#ifdef BARRETT_SMF_VALIDATE_ARGS
	bool validate_args(int argc, char** argv);
#endif

#ifdef BARRETT_SMF_CONFIGURE_PM
	bool configure_pm(int argc, char** argv, ::barrett::ProductManager& pm)
	{
	  ::barrett::SafetyModule* sm = pm.getSafetyModule();
	  bypassSafetyModule(sm);
	  ::barrett::SafetyModule::PendantState ps;

	  sm->getPendantState(&ps);
	  if (ps.pressedButton == ::barrett::SafetyModule::PendantState::ESTOP)
	    std::cout << "Please Release the WAM ESTOP to start the wam_node" << std::endl;
	  while (ps.pressedButton == ::barrett::SafetyModule::PendantState::ESTOP)
	  {
	    sm->getPendantState(&ps);
	    usleep(100000);
	  }
	  if(sm->getMode() == ::barrett::SafetyModule::ESTOP || sm->getMode() == ::barrett::SafetyModule::IDLE) {
	    sm->setMode(::barrett::SafetyModule::IDLE);
	    sm->waitForMode(::barrett::SafetyModule::IDLE, false);
	  }
	  return true;
	}
#endif

#ifndef BARRETT_SMF_NO_DECLARE
	template<size_t DOF> int proficio_main(int argc, char** argv, ::barrett::ProductManager& pm, ::barrett::systems::Wam<DOF>& wam, config);
#endif

#ifndef BARRETT_SMF_DONT_WAIT_FOR_SHIFT_ACTIVATE
#  define BARRETT_SMF_WAIT_FOR_SHIFT_ACTIVATE true
#else
#  define BARRETT_SMF_WAIT_FOR_SHIFT_ACTIVATE false
#endif

#ifndef BARRETT_SMF_DONT_PROMPT_ON_ZEROING
#  define BARRETT_SMF_PROMPT_ON_ZEROING true
#else
#  define BARRETT_SMF_PROMPT_ON_ZEROING false
#endif

#ifndef BARRETT_SMF_WAM_CONFIG_PATH
#  define BARRETT_SMF_WAM_CONFIG_PATH NULL
#endif


int main(int argc, char** argv) {
	// Give us pretty stack-traces when things die
	::barrett::installExceptionHandler();

try {

#ifdef BARRETT_SMF_VALIDATE_ARGS
	if ( !validate_args(argc, argv) ) {
		return 1;
	}
#endif

	::barrett::ProductManager pm;
#ifdef BARRETT_SMF_CONFIGURE_PM
	if ( !configure_pm(argc, argv, pm) ) {
		return 1;
	}
#endif

#ifndef NO_CONTROL_PENDANT
	pm.waitForWam(BARRETT_SMF_PROMPT_ON_ZEROING);
#endif
	pm.wakeAllPucks();

	// TODO(JH): Rehab Update implement and test
	if (pm.foundWam3()) {

		::barrett::systems::Wam<3>* wam = pm.getWam3(BARRETT_SMF_WAIT_FOR_SHIFT_ACTIVATE, BARRETT_SMF_WAM_CONFIG_PATH);
		BARRETT_UNITS_TEMPLATE_TYPEDEFS(3);
	#ifdef NO_CONTROL_PENDANT
 
  std::cout << "starting wam thread..." << std::flush;
  ::barrett::SafetyModule* sm = pm.getSafetyModule();
  for (int i=0; i<3; i++) {
    if(sm->getMode() != ::barrett::SafetyModule::ACTIVE){
      std::cout << " " << i << std::flush;
      barrett::systems::ExposedOutput<jt_type> zeroTorque;
      jt_type zero(0.0);
      zeroTorque.setValue(zero);
      barrett::systems::connect(zeroTorque.output,wam->input);
      usleep(50000);
      sm->setMode(barrett::SafetyModule::ACTIVE);
      if (i >= 2) {
        sm->waitForMode(barrett::SafetyModule::ACTIVE, false);
      } else {
        usleep(200000);
      }
      barrett::systems::disconnect(wam->input);
    }
  }
  std::cout << std::endl;
  wam->getLowLevelWam().getPuckGroup().setProperty(barrett::Puck::MODE, barrett::MotorPuck::MODE_TORQUE);

  /*
		::barrett::SafetyModule* sm = pm.getSafetyModule();
		if(sm->getMode() != ::barrett::SafetyModule::ACTIVE){
		   ::barrett::systems::ExposedOutput<jt_type> zeroTorque;
  		   jt_type zero(0.0);
  		   zeroTorque.setValue(zero);
  		   ::barrett::systems::connect(zeroTorque.output,wam->input);
  		   usleep(50000);

  		   pm.getSafetyModule()->setMode(::barrett::SafetyModule::ACTIVE);
  		   pm.getSafetyModule()->waitForMode(::barrett::SafetyModule::ACTIVE, false);
  		   ::barrett::systems::disconnect(wam->input);
		}
  		wam->getLowLevelWam().getPuckGroup().setProperty(::barrett::Puck::MODE, ::barrett::MotorPuck::MODE_TORQUE);
  */

	#endif
    		jp_type jpos = wam->getJointPositions();
		config side;
   		if(jpos[2] < 0) { // Left Config
      			side = LEFT;
    		} else if(jpos[2] > 0) { // Right Config
      			side = RIGHT;
    		}
		return proficio_main(argc, argv, pm, *pm.getWam3(BARRETT_SMF_WAIT_FOR_SHIFT_ACTIVATE, BARRETT_SMF_WAM_CONFIG_PATH), side);
	} else {
		printf(">>> ERROR: No Proficio was found. Perhaps you have found a bug in ProductManager::waitForWam().\n");
		return 0;
	}
} catch(std::runtime_error& e) {
		std::cout<<"ERROR: Could not instantiate the connection. Please check the logs "<<std::endl;

		return 0;
	}
}


#endif /* PROFICIO_STANDARD_MAIN_FUNCTION_H_ */
