/* collision_wall.h
 *
 * Given a separation distance between two robots in a bimanual system,
 * create haptic walls in the sagittal plane to prevent the two robots
 * from crossing each other.
 * 
 * This system applies forces to one robot only. To put safety walls with
 * both robots, two systems are needed.
 * 
 * I/O:
 * 		inputThis 		Cartesian position of the robot to be controlled by this system
 * 		inputOther 		Cartesian position of the other robot
 * 		output				Cartesian force applied to the controlled robot
 *
 * parameters:
 * 		separationDistance 	translation (meters) between the two robot world frame origins,
 * 												specified in the controlled robot's world frame.
 *
 * Note that inputThis and output can be connected directly to the 
 * controlled robot, but inputOther cannot be connected directly to the
 * other robot, since it is supervised by a different product manager. It
 * is assumed that the user of this system has created some thread-safe 
 * method of getting the other robot's position and (for example) assigning 
 * it to the value of an ExposedOutput system supervised by the same product 
 * manager as the controlled robot.
 * 
 *
 * Created by: ab
 * Date: 1/18/16
 * 
 */

#ifndef BARRETT_PROFICIO_COLLISION_WALL_H_
#define BARRETT_PROFICIO_COLLISION_WALL_H_


#include <barrett/systems.h>
#include <barrett/systems/abstract/system.h>
#include <barrett/units.h>

using namespace barrett;

template<size_t DOF>
class collisionWall : public systems::System {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);

public:
	Input<typename barrett::units::CartesianPosition::type> inputThis;
	Input<typename barrett::units::CartesianPosition::type> inputOther;
	Output<typename barrett::units::CartesianForce::type> output;

protected:
	typename Output<typename barrett::units::CartesianForce::type>::Value* outputValue;

public:
	collisionWall(const cp_type separationDistance_) :
		inputThis(this),
		inputOther(this),
		output(this,&outputValue),
		separationDistance(separationDistance_),
		error(0.0),
		force(0.0),
		outputData(0.0),
		Kp(500.0)
	{
		// define direction of wall based on robot config, which can be determined by sign of
		// y component of separationDistance.
		if (separationDistance[1] > 0) {
			// right config
			direction = -1;
		} else if (separationDistance[1] < 0) {
			// left config
			direction = 1;
		} else {
			// unknown. disable wall.
			direction = 0;
		}
	}

	~collisionWall() { this->mandatoryCleanUp(); }

protected:
	const cp_type separationDistance;
	double direction;
	cp_type error;
	cf_type force, outputData;
	const double Kp;

	virtual void operate() {
		error = inputThis.getValue() - inputOther.getValue() + separationDistance; // distance in meters between robot world frame origins
		double depth = error[1];
		if (direction*depth < 0) {
			force.setZero();
			force[1] = -1.0*Kp*depth;
		} else {
			force.setZero();
		}
		outputData = force;
		this->outputValue->setData(&outputData);
	}

private:
	DISALLOW_COPY_AND_ASSIGN(collisionWall);
};

#endif /* BARRETT_PROFICIO_COLLISION_WALL_H_ */
