/* haptic_bubble.h
 *
 * Given a separation distance between two robots in a bimanual system,
 * create haptic "bubbles" around the endpoints to prevent collision.
 * This could be used to place a bubble around another object as well,
 * as long as the position is specified relative to the correct origin
 * (assumed to be the origin of another robot with opposite configuration).
 * 
 * This system applies forces to one robot only. To put haptic bubbles
 * around both robots, two systems are needed.
 * 
 * I/O:
 * 		inputThis 		Cartesian position of the robot to be controlled by this system
 * 		inputOther 		Cartesian position of the other robot (or object) to avoid
 * 		output				Cartesian force applied to the controlled robot
 *
 * parameters:
 * 		separationDistance 	translation (meters) between the two robot world frame origins,
 * 												specified in the controlled robot's world frame.
 * 		bubbleRadius 				distance (meters) at which repulsive forces start to be applied
 * 												(measured between robot endpoint centers).
 *
 * Note that inputThis and output can be connected directly to the 
 * controlled robot, but inputOther cannot be connected directly to the
 * other robot, since it is supervised by a different product manager. It
 * is assumed that the user of this system has created some thread-safe 
 * method of getting the other robot's position and (for example) assigning 
 * it to the value of an ExposedOutput system supervised by the same product 
 * manager as the controlled robot.
 * 
 *
 * Created by: ab
 * Date: 1/14/16
 * 
 */

#ifndef BARRETT_PROFICIO_HAPTIC_BUBBLE_H_
#define BARRETT_PROFICIO_HAPTIC_BUBBLE_H


#include <barrett/systems.h>
#include <barrett/systems/abstract/system.h>
#include <barrett/units.h>

using namespace barrett;

template<size_t DOF>
class hapticBubble : public systems::System {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);

public:
	Input<typename barrett::units::CartesianPosition::type> inputThis;
	Input<typename barrett::units::CartesianPosition::type> inputOther;
	Output<typename barrett::units::CartesianForce::type> output;

protected:
	typename Output<typename barrett::units::CartesianForce::type>::Value* outputValue;

public:
	hapticBubble(const cp_type separationDistance_, const double bubbleRadius_) :
		inputThis(this),
		inputOther(this),
		output(this,&outputValue),
		separationDistance(separationDistance_),
		bubbleRadius(bubbleRadius_),
		error(0.0),
		force(0.0),
		outputData(0.0),
		Kp(500.0)
	{}

	~hapticBubble() { this->mandatoryCleanUp(); }

protected:
	const cp_type separationDistance;
	const double bubbleRadius;
	cp_type error;
	cf_type force, outputData;
	const double Kp;

	virtual void operate() {
		error = inputThis.getValue() - inputOther.getValue() + separationDistance; // distance in meters between robot world frame origins
		double mag = error.norm();
		if (mag < bubbleRadius) {
			force = Kp*(bubbleRadius - mag)*(error/mag);
		} else {
			force.setZero();
		}
		outputData = force;
		this->outputValue->setData(&outputData);
	}

private:
	DISALLOW_COPY_AND_ASSIGN(hapticBubble);
};

#endif /* BARRETT_PROFICIO_HAPTIC_BUBBLE_H_ */
