/*
	Copyright 2015 Barrett Technology <support@barrett.com>

	This file is part of libbarrett.

	This version of libbarrett is free software: you can redistribute it
	and/or modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	This version of libbarrett is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this version of libbarrett.  If not, see
	<http://www.gnu.org/licenses/>.

*/

/** Apply damping when joint velocity gets above a certain threshold.
 *
 * @file jv_sat.h
 * @date Oct 13, 2015
 * @author ab
 */

#include <barrett/systems.h>
#include <barrett/systems/abstract/system.h>
#include <barrett/units.h>

template<size_t DOF>
class jvSat : public barrett::systems::SingleIO<typename barrett::units::JointVelocities<DOF>::type, typename barrett::units::JointTorques<DOF>::type> {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);

public:
	jvSat(const v_type damping, const jv_type limits, const int mode = 1, const std::string& sysName = "jvSat") :
		barrett::systems::SingleIO<jv_type, jt_type>(sysName), dampingConstants(damping), velocityLimits(limits), multiplier(1.0), torqueLimit(35.0), mode(mode) {}

	virtual ~jvSat() { this->mandatoryCleanUp(); }

	void enable() {
		multiplier = 1.0;
	}

	void disable() {
		multiplier = 0.0;
	}

	void setDampingConst(v_type& val){
		dampingConstants = val;
	}

	void setJvLimit(jv_type& val){
		velocityLimits = val;
	}
protected:

	virtual void operate() {

		velocity = this->input.getValue();

		for (size_t i=0; i<DOF; i++){
			if (fabs(velocity[i]) > velocityLimits[i]) {
				if (mode == 1) {
					// linear damping
					double error = (velocity[i] - velocityLimits[i]*velocity[i]/fabs(velocity[i]));
					torque[i] = multiplier * -dampingConstants[i] * error;
				} else if (mode == 2) {
					// quadratic damping
					double error = (velocity[i] - velocityLimits[i]*velocity[i]/fabs(velocity[i]));
					torque[i] = multiplier * -dampingConstants[i] * error*fabs(error);
				} else {
					// output zero torque
					torque[i] = 0;
				}
				if (fabs(torque[i]) > torqueLimit) {
					torque[i] = torqueLimit*torque[i]/fabs(torque[i]);
				}
			} else {
				torque[i] = 0;
			}
		}
		this->outputValue->setData(&torque);

	}

	jv_type velocity;
	jt_type torque;
	v_type dampingConstants;
	jv_type velocityLimits;
	double multiplier;
	const double torqueLimit;
	const int mode; // 1 for linear damping, 2 for quadratic damping

private:
	DISALLOW_COPY_AND_ASSIGN(jvSat);
};
