#include <iostream>
#include <string>

#include <barrett/systems.h>
#include <barrett/systems/abstract/system.h>
#include <barrett/units.h>
#include <boost/tuple/tuple.hpp>

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Delaunay_triangulation_2.h>
#include <CGAL/natural_neighbor_coordinates_2.h>

typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef CGAL::Delaunay_triangulation_2<K> Delaunay_triangulation;
typedef std::vector< std::pair< K::Point_2, K::FT  > > Point_coordinate_vector;

template<size_t DOF>
class UserGravComp : public barrett::systems::System {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);

public:

	Input<typename barrett::units::JointPositions<DOF>::type> input;
	Output<typename barrett::units::JointTorques<DOF>::type> output;

protected:	typename Output<typename barrett::units::JointTorques<DOF>::type>::Value* outputValue;

public:
	explicit UserGravComp(const std::string& configFileName = "ugc_config.txt", const std::string& sysName = "UserGravComp") : barrett::systems::System(sysName), input(this), output(this, &outputValue) {
		// Get calibration data from config file.
		// Note that the order of calibration points is fixed such that points 2 through 5
		// are traveling clockwise around the edges of the convex hull and point 1 is on the
		// interior. This simplifies the process of projecting an external point onto the 
		// surface of the hull.
		// 		0: start position (approx 0.25, 0.00, 1.57)
		// 		1: q1 low, q2 low (arm up, out to side, towards user)
		// 		2: q1 low, q2 high (arm up, in front of user)
		// 		3: q1 high, q2 high (arm down, in front of user, away)
		// 		4: q1 high, q2 low (arm down, out to side)
		// 		5-14: whatever you want, or empty
		if (readConfig(calJoints,calTorques,configFileName)) {
		} else {
			std::cout << "Config file read failed.\n\n";
		}
		for (int i=1; i<5; i++) {
			calCenter += calJoints[i]/4;
		}
		torquePrevious << 0, 0, 0;
		outputData << 0, 0, 0;
		gain << 0, 0, 0;
		// create a Delaunay triangulation from the calibration data
		for (size_t i=0; i<calJoints.size(); i++) {
			dt.insert(K::Point_2(calJoints[i][0],calJoints[i][1]));
			//std::cout << "dt: inserted point " << calJoints.at(i) << std::endl;
		}
	}

	virtual ~UserGravComp() {this->mandatoryCleanUp();}

protected:

	std::vector<jp_type> calJoints;
	jp_type calCenter;
	jp_type gainCalSetpoint;
	jp_type gainCalPosition;
	std::vector<jt_type> calTorques;
	jt_type gainCalTorque;
	jt_type torquePrevious;
	jt_type outputData;
  Delaunay_triangulation dt;
	v_type gain;

public:

	void decrementGain() {
		v_type decrementVector(0.2);
		gain -= decrementVector;
		for (int i=0; i<3; i++) {
		if (gain[i] < 0)
			gain[i] = 0;
		}
		//return gain;
	}

	void incrementGain() {
		v_type incrementVector(0.2);
		gain += incrementVector;
		for (int i=0; i<3; i++) {
		if (gain[i] > 2)
			gain[i] = 2;
		}
		//return gain;
	}

	double getGain() { // This is invalid when each joint has its own gain
		return gain[0];
	}

	v_type resetGain() {
		gain << 1.0, 1.0, 1.0;
		return gain;
	}

	void setGainZero() {
		gain.setZero();
		//return gain;
	}

protected:

	int findPoint(const K::Point_2& point, const std::vector<jp_type> &pointList, const size_t length) {
		// searches the list for the point and returns the index when found.
		for (size_t i=0; i<length; i++) {
			if (fabs(point[0] - pointList[i][0]) < 0.001 && fabs(point[1] - pointList[i][1]) < 0.001) {
				return i;
			}
		}
		return -1; // none found
	}

	// Based on parseDoubles function from ex03.
	// Given a string of numbers separated by spaces, returns two 3-vectors of doubles.
	// The first vector is joint positions, the second is joint torques.
	bool parseDoubles(jp_type& x, jt_type& y, const std::string& str) {
		const char* cur = str.c_str();
		const char* next = cur;

		for (int i = 0; i < 3; ++i) {
			x[i] = strtod(cur, (char**) &next);
			if (cur == next) {
				return false;
			} else {
				cur = next;
			}
		}

		for (int i = 0; i < 3; ++i) {
			y[i] = strtod(cur, (char**) &next);
			if (cur == next) {
				return false;
			} else {
				cur = next;
			}
		}

		// Make sure there are no extra numbers in the string.
		double ignore = strtod(cur, (char**) &next);
		(void)ignore;  // Prevent unused variable warnings

		if (cur != next) {
			return false;
		}

		return true;
	}

	bool readConfig(std::vector<jp_type>& calJoints, std::vector<jt_type>& calTorques, std::string fname) {
		// Provide the absolute location of the file from which the constants are to be read
		std::ifstream myfile(fname.c_str());
		std::string line;
		size_t i=0;
		size_t size = 20; // max number of points to read
		jp_type dataJoints;
		dataJoints.setZero();
		jt_type dataTorques;
		dataTorques.setZero();
		if (myfile.is_open()) {
			while(getline(myfile,line) && i < size) {
				if (parseDoubles(dataJoints, dataTorques, line)) {
					//std::cout << "Got cal data " << dataJoints[0] << " " << dataJoints[1] << " " << dataJoints[2] << ", " << dataTorques[0] << " " << dataTorques[1] << " " << dataTorques[2] << std::endl;
				} else {
					std::cout << "ERROR: Parse cal data failed. " << line << std::endl;
					return false;
				}
				calJoints.push_back(dataJoints);
				calTorques.push_back(dataTorques);
				i++;
			}
			myfile.close();
		}
		jt_type dataZero;
		dataZero.setZero();
		calTorques.push_back(dataZero);
		return true;
	}


	bool interpolate(const jp_type& position, jt_type& torqueOut) {

		const K::Point_2 p(position[0],position[1]);
		const K::Vector_2 p_vec(p[0],p[1]);

		// empty vector for torque computation
		jt_type torque;
		torque.setZero();


		/* Procedure:
		 * (1)	Check whether point is inside or outside convex hull.
		 * 			Project the point onto each segment of the hull. Then check whether the point is
		 * 			inside or outside by multiplying by the normal pointing into the hull. If the point
		 * 			is outside for any segment, then it is outside the hull. Save distances between point
		 * 			and projections.
		 * (2)	If the point is outside for any segment, check whether its projection is on the
		 * 			segment. If yes, then the projection is the closest point on the hull. Perturb the
		 * 			point towards the hull center and calculate the nn coordinates. The perturbation
		 * 			prevents trying to calculate nn coordinates for a point too close to the edge, which
		 * 			takes a long time and sometimes results in inf or nan weights.
		 * (2a) Use nn coordinates to complete the interpolation. The nn coordinates stored in 
		 * 			coords.second are the weights to use for the corresponding points in coords.first. 
		 * 			Those weights need to be normalized by the sum of weights stored in result.second. 
		 * 			Since the output in coords is not in the same order as the original list of points, 
		 * 			search the list for the points in coords, in order, and pull out the appropriate 
		 * 			torque to multiply by the listed nn coordinate.
		 * (3)	If the point is inside the hull, make sure it is some minimum distance from all
		 * 			segments. If not, perturb it a small amount towards the hull center, again to prevent
		 * 			long calculation times and inf/nan weights. Then calculate the nn coordinates.
		 * (3a) Use nn coordinates to complete the interpolation as in (2a).
		 * (4)	If the point is not found in an edge section, then it must be in a corner section.
		 * 			Calculate the distance to each corner. The closest corner is the closest point on the
		 * 			surface of the hull. Use the calibration torque from that point.
		 *
		 * Assumptions:
		 * (1)	Points were collected in the specified order:
		 * 				1: start position (approx 0.25, 0.00, 1.57)
		 * 				2: q1 low, q2 low (arm up, out to side, towards user)
		 * 				3: q1 low, q2 high (arm up, in front of user)
		 * 				4: q1 high, q2 high (arm down, in front of user, away)
		 * 				5: q1 high, q2 low (arm down, out to side)
		 * 				6-20: whatever you want, or empty
		 * (2)	Points are at least 0.003 apart.
		 * (3)	Start point is set at (0.2647750, -0.0058371, 1.5773100), or at least 0.002 away from (0,0).
		 */

		//coordinate computation
		const K::Point_2 center(calCenter[0],calCenter[1]);
		bool pointFound = false;
		bool outside = false;
		bool nearEdge = false;
		bool success = true;

		// Check if point is inside or outside convex hull created by calibration points.

		for (int seg=1; seg<5; seg++) {

			// define the line segment
			K::Point_2 p1(calJoints[seg][0],calJoints[seg][1]);
			double p2x, p2y;
			if (seg<4) {
				p2x = calJoints[seg+1][0];
				p2y = calJoints[seg+1][1];
			} else {
				p2x = calJoints[1][0];
				p2y = calJoints[1][1];
			}
			K::Point_2 p2(p2x,p2y);

			// get a unit vector in the direction of the line segment
			K::Vector_2 segment(p1,p2);
			segment = segment/sqrt(segment.squared_length());
			// project the point of interest onto that vector
			K::Point_2 proj = p1 - ((p1-p)*segment)*segment;
			// find the normal pointing into the hull
			K::Vector_2 normal(segment[1],-segment[0]);
			// check whether point is outside the hull on this side
			if (normal*(proj-p) > 0) {
				// point is outside this segment. check if the projection is on the segment.
				outside = true;
				if ((proj[0] >= std::min(p1[0],p2[0])) && (proj[0] <= std::max(p1[0],p2[0])) && (proj[1] >= std::min(p1[1],p2[1])) && (proj[1] <= std::max(p1[1],p2[1]))) {
					// This is the section containing the point.
					// Perturb the projection towards center.
					K::Vector_2 toCenter(proj,center);
					proj = proj + 0.000001*toCenter;
					// Get the nn coordinates
					Point_coordinate_vector coords;
					CGAL::Triple< std::back_insert_iterator<Point_coordinate_vector>, K::FT, bool> result
						= CGAL::natural_neighbor_coordinates_2(dt, proj, std::back_inserter(coords));
					// check success
					if (result.third) {
						// Calculate the torque
						for (size_t i=0; i<coords.size(); i++) {
							int index = findPoint(coords[i].first,calJoints,calJoints.size());
							if (index < 0) {
								index = calTorques.size(); // zero torques
							}
							for (int j=0; j<3; j++) {
								torque[j] += coords[i].second*calTorques[index][j]/result.second;
							}
						}
					} else {
						// Failed. Use torque from previous timestep.
						success = false;
						std::cout << "ERROR: NN coordinate calculation failed.\n";
					}
					// don't bother checking the other edge sections
					pointFound = true;
					break;
				}
			} else {
				// point is inside this segment. check distance to edge.
				if (sqrt((proj-p).squared_length()) < 0.000001) {
					nearEdge = true;
				}
			}
		}

		if (!outside) {
			// point is inside the convex hull.
			// check distances to edges.
			K::Point_2 p_new(p);
			if (nearEdge) {
				// perturb towards center
				K::Vector_2 toCenter(p,center);
				p_new = p + 0.000001*toCenter;
			}
			// Get the nn coordinates
			Point_coordinate_vector coords;
			CGAL::Triple< std::back_insert_iterator<Point_coordinate_vector>, K::FT, bool> result
				= CGAL::natural_neighbor_coordinates_2(dt, p_new, std::back_inserter(coords));
			// check success
			if (result.third) {
				// Calculate the torque
				for (size_t i=0; i<coords.size(); i++) {
					int index = findPoint(coords[i].first,calJoints,calJoints.size());
					if (index < 0) {
						index = calTorques.size(); // zero torques
					}
					for (int j=0; j<3; j++) {
						torque[j] += coords[i].second*calTorques[index][j]/result.second;
					}
				}
			} else {
				// Failed. Use torque from previous timestep.
				success = false;
				std::cout << "ERROR: NN coordinate calculation failed.\n";
			}
			pointFound = true;
		} else if (!pointFound) {
			// point is outside the convex hull but not in an edge section.
			// point must be in a corner section.
			// find the closest corner point and use that.
			int min_i = 1;
			double min_dist = sqrt(pow(p[0]-calJoints[1][0],2) + pow(p[1]-calJoints[1][1],2));
			double dist;
			for (int i=2; i<5; i++) {
				dist = sqrt(pow(p[0]-calJoints[i][0],2) + pow(p[1]-calJoints[i][1],2));
				if (dist<min_dist) {
					min_dist = dist;
					min_i = i;
				}
			}
			for (int i=0;i<3;i++) {
				torque[i] = calTorques[min_i][i];
			}
		}

		if (success) {
			torqueOut = torque;
		}
		return success;
	}

	virtual void operate() {

		// get current robot position from input
		jp_type position = this->input.getValue();
		jt_type torque(torquePrevious);

		// Do the user gravity compensation calculation. If calculation is successful, the value in
		// torque is overwritten.  If calculation fails, torque is unchanged from previous value.
		if(interpolate(position,torque)) {
			// Multiply by gain for the current user.
			for (size_t i=0; i<DOF; i++) {
				torque[i] = gain[i]*torque[i];
			}
		}

		// set torque output
		{
			BARRETT_SCOPED_LOCK(this->getEmMutex());
			outputData = torque;
			this->outputValue->setData(&outputData);
		}

		// Save torque output value.
		torquePrevious = torque;
	}

private:
	DISALLOW_COPY_AND_ASSIGN(UserGravComp);
};


