/*
 * ball_popping.cpp
 *
 * Created on: Apr 04, 2015
 *      Author: Hariharasudan Malaichamee
 */
#include <boost/bind.hpp>
#include <boost/tuple/tuple.hpp>

#include <barrett/exception.h>
#include <barrett/units.h>
#include <barrett/systems.h>
#include <barrett/products/product_manager.h>
#include <barrett/config.h>
#include <proficio/systems/utilities.h>

#include <signal.h>
#include <fstream>
#include <math.h>
#include <syslog.h>
#include "ball_popping.h"

#define BARRETT_SMF_VALIDATE_ARGS
#define NO_CONTROL_PENDANT

#include <proficio/standard_proficio_main.h>

BARRETT_UNITS_FIXED_SIZE_TYPEDEFS;
BARRETT_UNITS_TYPEDEFS(15);

char* remoteHost = NULL;
double kp = 3e3;
double kd = 3e1;
double segment_s;

bool validate_args(int argc, char** argv) {
  switch (argc) {
    case 3:
      remoteHost = argv[1];
      segment_s = atof(argv[2]);
      break;

    default:
      printf("Usage: %s <otherip> <session_time> \n", argv[0]);
      return false;
      break;
  }

  printf("Gains: kp = %f; kd = %f\n", kp, kd);
  return true;
}

double restart = 0;
bool game_exit = false;
double TIMER = 1;
bool leftConfig;
namespace ball_popping {
bool exit_ = false;
const double MAGFIELD = 0.02;  // The distance of magnetic field around the floor
double MAGFIELD_BALL = 0.05;  // the distance of the magnetic field around the ball
const double FLOOR_W = 0.01;  // Width of the floor
double FLOOR = 0.05;  // Floor's absolute Z
double TOOL_R = 0.05;  // Radius of the end point. should match with that in the python client

// Python Client's workspace. Should match with that in the python client
double client_minX = -1;
double client_maxX = 0.5;
double client_minY = -1.4;
double client_maxY = 0.6;
double client_minZ = 0.06; // Floor + wall size
double client_maxZ = 1.55;

// The actual workspace's original range which would be expanded upon calibration
double min_x = 0.25;
double max_x = 0.35;
double min_y = 0.15;
double max_y = 0.25;
double min_z = 0.15;
double max_z = 0.25;

double j1_max = 0.55;
double j1_min = -1.0;
double j2_max = 0.95;
double j2_min = -0.95;

double j3L_max = -0.39;
double j3L_min = -2.99;

double j3R_max = 2.99;
double j3R_min = 0.37;

v_type msg_tmp;

barrett::systems::ExposedOutput<v_type> message;

cf_type scale(boost::tuple<cf_type, double> t) {
  return t.get<0>() * t.get<1>();
}

/*
 * Function to convert the coordinates from the the python client's frame to the calibrated frame
 */
inline cp_type descale(cp_type val) {
  val[0] = min_x
      + ((val[0] - client_minX) * static_cast<double>(max_x - min_x)
          / (client_maxX - client_minX));
  val[1] = min_y
      + ((val[1] - client_minY) * static_cast<double>(max_y - min_y)
          / (client_maxY - client_minY));
  val[2] = min_z
      + ((val[2] - client_minZ) * static_cast<double>(max_z - min_z)
          / (client_maxZ - client_minZ));

  return val;
}

/*
 * Function to convert the coordinates from the calibrated frame to the python client's frame
 */
inline cp_type scale(cp_type val) {
  val[0] = client_minX
      + ((val[0] - min_x) * static_cast<double>(client_maxX - client_minX)
          / (max_x - min_x));
  val[1] = client_minY
      + ((val[1] - min_y) * static_cast<double>(client_maxY - client_minY)
          / (max_y - min_y));
  val[2] = client_minZ
      + ((val[2] - min_z) * static_cast<double>(client_maxZ - client_minZ)
          / (max_z - min_z));

  return val;
}

/* check if the sum of the radius of the two spheres are
 * greater than the euclidean distance between their centers
 */
inline bool reached(cp_type sp, double sp_r, cp_type gl, double gl_r) {
  return ((sp - gl).norm() <= (sp_r + gl_r));
}

template <size_t DOF>
class SolidHapticBall : public barrett::systems::HapticBall {
  BARRETT_UNITS_FIXED_SIZE_TYPEDEFS;

protected:
  cp_type toolCenter, ballCenter, proxyCenter, shrinkSphereOrigin,
  shrinkSphere_ctr, prev_ballCenter;
  cf_type cfSum, direction;
  cv_type wamVelocity, ballVelocity;
  double stiffness, damping, stiff, damp, scale_x,
  shrinkSphere_r, counter, ss_kp, ss_kd, mag_ball_kp, max_kp, min_kp, max_kd,
  min_kd, MAG_FLOOR, speed, game_mode, timer, force_delay;
  bool prevState, posChanged, inside, cbt_init, attached, wait,
  game_pause, next_seg, it_wait;
  int score;
  UserGravComp<DOF> * gc;

  void initialize(const cp_type& center){
	  ballCenter = center;
	  stiffness = 1000;
	  damping = 2;
	  shrinkSphere_r = 1.5;
	  counter = 0;
	  ss_kp = 220;
	  ss_kd = 1;
	  mag_ball_kp = 1400;
	  max_kp = 700;
	  min_kp = 50;
	  max_kd = 4;
	  min_kd = 1;
	  speed = 0.0005;
	  game_mode = -1;
	  prevState = false;
	  posChanged = false;
	  calibrate = false;
	  inside = false;
	  cbt_init = false;
	  attached = false;
	  wait = false;
	  game_pause = false;
	  score = 0;
	  new_pos = false;
	  next_seg = true;
	  BALLR = 0.5;
	  it_wait = false;
	  force_delay = 0.03;
	  shrinkSphere_ctr.setZero();
	  shrinkSphereOrigin.setZero();
	  ballVelocity.setZero();
	  prev_ballCenter.setZero();
	  counter = segment_s;
	  MAG_FLOOR = FLOOR + FLOOR_W;
	  timer = TIMER;
	  scale_x = (toolCenter[0] - client_minX) * (1/(client_maxX - client_minX));
	  stiff = min_kp + ((scale_x) * (max_kp - min_kp));
	  damp = min_kd + ((scale_x) * (max_kd - min_kd));
  }
  // IO
public:
  barrett::systems::System::Output<cf_type> output;
  barrett::systems::System::Input<boost::tuple<double, double, double, double, double, double, double> >
  ball_input;
  barrett::systems::System::Input<cv_type> wamVel_input;
  bool calibrate, new_pos;
  cp_type cart_pos;
  double BALLR;

protected:
  barrett::systems::System::Output<cf_type>::Value* outputValue;

public:
  SolidHapticBall(UserGravComp<DOF> &ugc, const cp_type& center, double radius, const
      std::string& sysName = "HapticBall"):
  HapticBall(center, radius, sysName), gc(&ugc), output(this, &outputValue), ball_input(this),
  wamVel_input(this) {
	  initialize(center);
	    readConfig();
  }
  virtual ~SolidHapticBall() {mandatoryCleanUp();}

  void readConfig() {
    char fname[] = "config.txt";
    // Provide the absolute location of the file from which the constants are to be read
    std::ifstream myfile(fname);
    std::string line;
    size_t i=0;
    size_t size = 4;
    std::vector<double> a(size, 0);
    if (myfile.is_open()) {
      while(getline(myfile,line) && i < size) {
        a[i] = atof(line.c_str());
        i++;
      }
      myfile.close();
    }
    ss_kp = a[0];
    ss_kd = a[1];
    mag_ball_kp = a[2];
  }

  inline bool inEllipsoid(cp_type pos){
	  double r1, r2, r3;
	  r1 = (max_x - min_x) / 2;
	  r2 = (max_y - min_y) / 2;
	  r3 = (max_z - min_z) / 2;
	  cp_type center;
	  center << min_x + r1, min_y + r2, min_z + r3;
	  double dist = pow((pos[0] - center[0])/r1,2) + pow((pos[1] - center[1])/r2,2) + pow((pos[2] - center[2])/r3,2);
	  if (dist < 1)
	  	  return true;
      return false;
  }

  inline bool isDistanced(cp_type pos){
	 pos = scale(pos);
	 if ((pos - prev_ballCenter).norm() >= 0.75)
	 	 return true;
     return false;
 }

protected:
  virtual void operate() {
    // Read the Inputs
    toolCenter = input.getValue();
    wamVelocity = wamVel_input.getValue();
    const double& ballRadius = ball_input.getValue().get<0>();
    BALLR = ballRadius;

    /* Indicates if the Sphere Center and the Tool Position are to be scaled to the Client's workspace or not
     * 1 = Scale the Tool Center. This scaling would be done only when no forces are acting on the Robot
     */
    double scl = ball_input.getValue().get<1>();

    /* Indicates the User's gravity compensation gain
     * 1 = Increment the gain
     * 2 = Decrement the gain
     * 3 = Set the gain as 0
     */
    double gain = ball_input.getValue().get<2>();

    /* Indicates the Forces that needs to be activated based on the part of the game that is activated
     * 0 = No forces. This is generally active till the user moves the tool position into a safety sphere to activate the forces
     * 1 = Just the floor and the wall close to the user are active. This is before the workspace calibration
     * 2 = The front wall and the floor are active. This is while calibrating the workspace.
     * 3 = The actual game. The sphere with magnetic force, 6 walls, shrinking sphere are all active.
     * 4 = New session of the game. Resets the parameters and sets the game mode back to 3.
     * 5 = Quit the game
     */
    game_mode = ball_input.getValue().get<4>();
    cfSum.setZero(); // Reset the forces at each iteration
    if(gain == 1){
    	gc->incrementGain();
    }
    else if(gain == 2){
    	gc->decrementGain();
    }
    else if(gain == 3){
    	gc->setGainZero();
    }
    if(game_mode == 0){ // This game mode enters into none of the loops below that generates forces
    	if (scl == 1) // This combination is triggered if we need to have 0 forces after defining a workspace.
    		toolCenter = scale(toolCenter);
    }
    if(game_mode == 1){ // Just the magnetic Floor
       	initialize(ballCenter);
        // Floor forces
        if (toolCenter[2] < MAG_FLOOR + (TOOL_R)) {
         proxyCenter[2] = MAG_FLOOR + (TOOL_R);
          cfSum[2] += ((2 + (scale_x * 575)) * (proxyCenter[2] - toolCenter[2]));
        }
        // Front Wall forces
        if(toolCenter[0] > client_maxX - TOOL_R) {
          proxyCenter[0] = client_maxX - TOOL_R;
          cfSum[0] += (75 * (proxyCenter[0] - toolCenter[0]));
        }
    }
    if (game_mode == 2 ) { // Workspace calibration
    	if(!cbt_init){ // This condition is to initialize the 3d workspace around the current position of the Wam's end point
    	      max_x = toolCenter[0] + 0.05;
    	      min_x = toolCenter[0] - 0.05;
    	      max_y = toolCenter[1] + 0.05;
    	      min_y = toolCenter[1] - 0.05;
    	      min_z = MAG_FLOOR;
    	      max_z = toolCenter[2] + 0.05;
    	      cbt_init = true;
    	}
    	else{ // If the workspace is already initialized, start exapnding them
      		  min_x = toolCenter[0] < min_x ? toolCenter[0]:min_x;
      		  if (toolCenter[0] > max_x && max_x < client_maxX){ // Limit the workspace not to extend beyond the front wall
      		    max_x = toolCenter[0];
      		  }
    		    if(leftConfig){ // Left Configuration Robot
    	 		      if (toolCenter[1] < min_y){ // This change is to account for the kinematic constraints in the left trip when the third joint is closed to its 0
    	 		       	min_y = toolCenter[1];
    	 		      }
    	 		      max_y = toolCenter[1] > max_y ? toolCenter[1]:max_y;
    		     }
    		    else{ // Right Configuration Robot

    	 		        min_y = toolCenter[1] < min_y ? toolCenter[1]:min_y;
    	 		        if (toolCenter[1] > max_y && toolCenter[1] < 0.2){
    	 		        	max_y = toolCenter[1];
    	 		        }
    		     }
    		     max_z = toolCenter[2] > max_z ? toolCenter[2]:max_z;

    	}
    	  // Floor forces
        if (toolCenter[2] < MAG_FLOOR + (TOOL_R)) {
          proxyCenter[2] = MAG_FLOOR + (TOOL_R);
          cfSum[2] += ((2 + (scale_x * 575)) * (proxyCenter[2] - toolCenter[2]));
        }
        // Front Wall forces
        if(toolCenter[0] > client_maxX - TOOL_R) {
          proxyCenter[0] = client_maxX - TOOL_R;
          cfSum[0] += (75 * (proxyCenter[0] - toolCenter[0]));
        }

    }

    if(game_mode == 4){ // New Session
      // Reset the game parameters
    	next_seg = true;
    	counter = segment_s;
	    score = 0;
	    // Reset the mode to game
      game_mode = 3;
    }
    if(game_mode == 3){ // The game
        /* The rate at which the shrinking sphere's radius reduces at each loop
         * Possible options are 0.001, 0.0005, 0.00025
         */
        speed = ball_input.getValue().get<3>();
        /* The time taken for the onset of the shrinking sphere to come in contact with the tool position. Defines the edge offset of starting radius of the shrinking sphere
         * Possible options are 1, 0.5, 0.1
         */
        force_delay = ball_input.getValue().get<5>();
        /* The force with which the shrinking sphere pushes the tool position inside. Defines the gains used in the calculation.
         * Possible options are 500 and 0
         */
        ss_kp = ball_input.getValue().get<6>();
    	if(next_seg && !it_wait) { // After attaching to a sphere and between the time a new sphere appears
    		if (new_pos){ // If the main thread has generated a new random position for the sphere center
    			ballCenter = cart_pos;
    	    		ballCenter = scale(ballCenter);
    			toolCenter = scale(toolCenter);
    			attached = false;
    			new_pos = false;
    			next_seg = false; // Reset the parameters
    		  	wait = false;
    		  	shrinkSphere_r = (ballCenter - toolCenter).norm() + force_delay;
    	    		toolCenter = descale(toolCenter);
    		}
    		else {
    			wait = true; // Pause the game till we get the new parameters
			counter = counter - 0.002;
    			// counter = counter + 0.002;
                       // game_pause = true;
    		}
    	}
        toolCenter = scale(toolCenter);
    	if(!wait){ // If the game is not waiting decrease the timer
    		counter = counter - 0.002;

    	}
      game_pause = counter > 0 ? false : true; // Once the timer hits 0 pause the game. This marks the new session
      shrinkSphere_ctr = ballCenter;
      // shrink the radius of the sphere
      shrinkSphere_r = (shrinkSphere_r > 0) ? (shrinkSphere_r - speed) : 0;
      // Update the ball's parameters
      this->setCenter(ballCenter);
      this->setRadius(ballRadius);

      // Find the location of the tool's center relative to the ball
      error = toolCenter - ballCenter;
      double mag = error.norm();// Distance between the tool and the ball
      double depth = mag - r;// The distance at which the tool resides with respect to the ball's radius
      // Unit vector from the ball's center towards the tool's center
      direction = error / mag;
      proxyCenter = toolCenter;
      // If the tool is outside and within the magnetic field
      if (depth <= MAGFIELD + TOOL_R && depth > 0 && !game_pause && !wait && !it_wait) {
        proxyCenter = ballCenter + ((ballRadius + TOOL_R) * direction);  // The position along the direction of the tool on the surface of the ball

        cfSum[0] += mag_ball_kp * (proxyCenter[0] - toolCenter[0]);
        cfSum[1] += mag_ball_kp * (proxyCenter[1] - toolCenter[1]);
        cfSum[2] += mag_ball_kp * (proxyCenter[2] - toolCenter[2]);
      }
      if ( depth <= TOOL_R && !wait && !game_pause && !it_wait) {  // If the tool is going inside the surface of the ball
        attached = true; // This segment has ended
        next_seg = true; // It is time to reset the ball center to a new position and start it again
        it_wait = true;
        score ++;
      }
      // Constructing the floor on either sides
      // Back Wall
      if(toolCenter[0] < client_minX + TOOL_R) {
        proxyCenter[0] = client_minX + TOOL_R;
        cfSum[0] += (75 * (proxyCenter[0] - toolCenter[0]));
      }
      // Front Wall
      if(toolCenter[0] > client_maxX - TOOL_R) {
        proxyCenter[0] = client_maxX - TOOL_R;
        cfSum[0] += (75 * (proxyCenter[0] - toolCenter[0]));
      }
      // Left Wall
      if(toolCenter[1] < client_minY + TOOL_R + 0.03) {
        proxyCenter[1] = client_minY + TOOL_R + 0.03;
        cfSum[1] += ((2 + (scale_x * 75)) * (proxyCenter[1] - toolCenter[1]));
      }
      // Right Wall
      if(toolCenter[1] > client_maxY - TOOL_R - 0.03) {
        proxyCenter[1] = client_maxY - TOOL_R - 0.03;
        cfSum[1] += ((2 + (scale_x * 75)) * (proxyCenter[1] - toolCenter[1]));
      }
      // Top Wall
       if(toolCenter[2] > client_maxZ - TOOL_R - 0.03) {
         proxyCenter[2] = client_maxZ - TOOL_R - 0.03;
         cfSum[2] += ((2 + (scale_x * 75)) * (proxyCenter[2] - toolCenter[2]));
       }
      if ( (shrinkSphere_r - mag) < (TOOL_R) &&
          (shrinkSphere_r - mag) > 0 && !game_pause && !wait && !it_wait) {  // The tool is touching the shrinking sphere's surface from inside

        // Apply force 0 when just touching or inside sphereRad
        // Apply increasing force when between touching and proxy center
        // Apply large force when proxy center is at sphereRad
        // Apply force 0 when proxy center is beyond sphereRad
        proxyCenter = ballCenter + (shrinkSphere_r - TOOL_R) * direction;// On the inner surface of the shrinking sphere
        cfSum[0] += ss_kp * (proxyCenter[0] - toolCenter[0]);
        cfSum[1] += ss_kp * (proxyCenter[1] - toolCenter[1]);
        cfSum[2] += ss_kp * (proxyCenter[2] - toolCenter[2]);
      }

      // After each ball is reached wait for TIMER seconds before starting again
      if (it_wait)
      	timer = timer - 0.002;
      if (timer <= 0){
          it_wait = false;
          timer = TIMER;
      }
      prev_ballCenter = ballCenter;

      // Floor calculations
      if (toolCenter[2] < MAG_FLOOR + (TOOL_R)) {
       proxyCenter[2] = MAG_FLOOR + (TOOL_R);
        cfSum[2] += ((2 + (scale_x * 175)) * (proxyCenter[2] - toolCenter[2]));
      }
    }
    // If exit is signalled set the global variable as true
    else if(game_mode == 5)
    	game_exit = true;

    if (restart == 1) {  // If a torque/velocity fault has occurred, reset the forces and velocities
      cfSum.setZero();
      attached = false;
      game_pause = false;
      wait = false;
      score = 0;
      counter = segment_s;
      restart = 0;
    }
double gcal_gain = gc->getGain(); // Read the current User's gravity calibration gain
    // Convert the ball and the tool's positions to the client's frame
    /*
     *  Update the message with the wam position, ball position and
     *   an indication of whether the ball is attached
     */
    for (int j = 0; j < 15; j++) {
      if (j < 3)
      msg_tmp[j] = toolCenter[j];
      else if (j < 4)
      msg_tmp[j] = static_cast<double>(attached);
      else if (j < 5)
      msg_tmp[j] = shrinkSphere_r;
      else if (j < 6)
      msg_tmp[j] = score;
      else if (j < 7)
      msg_tmp[j] = static_cast<double>(game_pause);
      else if(j < 10)
      msg_tmp[j] = ballCenter[j-7];
      else if(j < 11)
      msg_tmp[j] = counter;
      else if(j < 12)
      msg_tmp[j] = static_cast<double>(it_wait);
      else if(j < 13)
      msg_tmp[j] = gcal_gain;
      else if(j < 14)
      msg_tmp[j] = static_cast<double>(leftConfig);
      else if(j < 15)
      msg_tmp[j] = static_cast<double>(leftConfig);
    }
    outputValue->setData(&cfSum);
    message.setValue(msg_tmp);
  }
};

template<size_t DOF>
void init(barrett::systems::Wam<DOF>& wam) {
  cp_type tp = wam.getToolPosition();
  cp_type center;
  center << 0.35, 0.6, 0.0;
  for (int j = 0; j < 7; j++) {
    if (j < 3)
      msg_tmp[j] = tp[j];
    else if (j < 6)
      msg_tmp[j] = center[j - 3];  // Indicates that the ball is not attached to the tool
    else
      msg_tmp[j] = 0;
  }
}

void exit_pgm(int signum)
{
  game_exit = true;
  return;
}
}

template<size_t DOF>
int proficio_main(int argc, char** argv, barrett::ProductManager& pm,
             barrett::systems::Wam<DOF>& wam, config side) {
  BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);
wam.gravityCompensate();
  barrett::SafetyModule* sm = pm.getSafetyModule();
  barrett::SafetyModule::PendantState ps;
  sm->getPendantState(&ps);
  if (DOF != 3){
	  printf("This program is supported only for the Proficio robot");
	  return 0;
  }
signal(SIGINT, ball_popping::exit_pgm);
    signal(SIGTERM, ball_popping::exit_pgm);
    signal(SIGKILL, ball_popping::exit_pgm);

  std::ofstream startFlag;
  startFlag.open("/tmp/bpstart");
  startFlag.close();

  std::string fname="calibration_data/wam3/";
  cp_type center;
  center << 0.4, -0.3, 0;
  cv_type freq;
  freq << 20, 20, 20;
  //Read the kinematics
  double j1_Max, j1_Min, j2_Max, j2_Min, j3_Max, j3_Min;

  j1_Max = ball_popping::j1_max;
  j1_Min = ball_popping::j1_min;

  j2_Max = ball_popping::j2_max;
  j2_Min = ball_popping::j2_min;

  if(side == LEFT) { // Left Config
	  leftConfig = true;
	  fname = fname+"LeftConfig.txt";
	  j3_Max = ball_popping::j3L_max;
	  j3_Min = ball_popping::j3L_min;
  }
  else if(side == RIGHT) {
	  leftConfig = false;
	  fname = fname+"RightConfig.txt";
	  j3_Max = ball_popping::j3R_max;
	  j3_Min = ball_popping::j3R_min;
  }
  else {
	  printf("No known robot found ");
	  return 0;
  }
  UserGravComp<DOF> gc(barrett::EtcPathRelative(fname).c_str());
  jp_type rand_ang;

  struct bt_kinematics * kin;
  config_setting_t * world;

  libconfig::Setting& wamSetting = pm.getConfig().lookup(pm.getWamDefaultConfigPath());
  bt_kinematics_create(&kin, wamSetting["kinematics"].getCSetting(), DOF);
  world = config_setting_get_member(wamSetting["kinematics"].getCSetting(),"world_to_base");

  ForwardKin<DOF> FwKin(kin, world);

  jp_type jp;
  jp << 0, 0, -1;
  cp_type cp;
  cp << -0.00207499, 0.288004, 0.319999;

  // instantiate Systems
  // Invert the X,Y axes for the Proficio and add an offset
  barrett::systems::modXYZ<cp_type> modcp;
  modcp.negX();
  modcp.negY();
  modcp.xOffset(1);
  // Invert the cartesian forces and the velocities for the Proficio
  barrett::systems::modXYZ<cf_type> modforce;
  modforce.negX();
  modforce.negY();
  barrett::systems::modXYZ<cv_type> invvel;
  invvel.negX();
  invvel.negY();

  NetworkHaptics nh(pm.getExecutionManager(), remoteHost, center);
  barrett::systems::FirstOrderFilter<cv_type> fof;
  fof.setLowPass(freq);
  ball_popping::init(wam);
  ball_popping::SolidHapticBall<DOF> shb(gc, center, 0.12);
  barrett::systems::ToolForceToJointTorques<DOF> cf_tf2jt;
  barrett::systems::Summer<jt_type, 3> jtSum("+++");
  ball_popping::message.setValue(ball_popping::msg_tmp);

  // Joint Torques Saturation
  jt_type jtLimits(35.0);
  jtSat<DOF> jtsat(jtLimits);

  // Joint Velocities Saturation
  v_type dampingConstants(20.0);
  dampingConstants[2] = 10.0;
  dampingConstants[0] = 30.0;
  pm.getSafetyModule()->setVelocityLimit(1.5);
  jv_type velocityLimits(1.4);
  jvSat<DOF> velsat(dampingConstants,velocityLimits);
  jv_type jvFiltFreq(20.0);
  barrett::systems::FirstOrderFilter<jv_type> jvFilter;
  jvFilter.setLowPass(jvFiltFreq);
  
  barrett::systems::connect(wam.jvOutput, jvFilter.input);
  barrett::systems::connect(jvFilter.output, velsat.input);

  barrett::systems::forceConnect(ball_popping::message.output, nh.input);
  barrett::systems::connect(nh.output, shb.ball_input);
  barrett::systems::connect(wam.jpOutput, gc.input);

  if(DOF == 3) {
    barrett::systems::connect(wam.toolPosition.output, modcp.input);
    barrett::systems::connect(modcp.output, shb.input);

    barrett::systems::connect(shb.output, modforce.input);
    barrett::systems::connect(modforce.output, cf_tf2jt.input);

    barrett::systems::connect(wam.toolVelocity.output, invvel.input);
    barrett::systems::connect(invvel.output, fof.input);
  } else {
    barrett::systems::connect(wam.toolPosition.output, shb.input);
    barrett::systems::connect(wam.toolVelocity.output, fof.input);
    barrett::systems::connect(shb.output, cf_tf2jt.input);
  }

  barrett::systems::connect(wam.kinematicsBase.kinOutput, cf_tf2jt.kinInput); 
  barrett::systems::connect(fof.output, shb.wamVel_input);
  barrett::systems::connect(cf_tf2jt.output, jtSum.getInput(0));
  barrett::systems::connect(gc.output, jtSum.getInput(1));
  barrett::systems::connect(velsat.output, jtSum.getInput(2));
  barrett::systems::connect(jtSum.output, jtsat.input);

  //Modify safety limits
  pm.getSafetyModule()->setTorqueLimit(4);
  pm.getSafetyModule()->setVelocityLimit(2);
  barrett::systems::connect(jtsat.output, wam.input);
  while (true) {
    
    wam.idle();
    barrett::btsleep(0.02);
    while (true) {
  	sm->getPendantState(&ps);
	// Adding this will cause the new position generation to delay thereby making the program to wait for a new ball
	// However, without the delay the pendant state will not be updated, hence the program will not quit on Estop
	// barrett::btsleep(0.02);
    	// Generates random joint angles within the joint range and does the forward kinematics for them
   	  rand_ang[0] = (j1_Max - j1_Min) * ( (double)rand() / (double)RAND_MAX ) + j1_Min;
   	  rand_ang[1] = (j2_Max - j2_Min) * ( (double)rand() / (double)RAND_MAX ) + j2_Min;
   	  rand_ang[2] = (j3_Max - j3_Min) * ( (double)rand() / (double)RAND_MAX ) + j3_Min;
    	  shb.cart_pos = FwKin.findFwKinPosition(rand_ang);
    	  shb.cart_pos[0] = - shb.cart_pos[0];
    	  shb.cart_pos[1] = - shb.cart_pos[1];
    	  shb.cart_pos[0] = shb.cart_pos[0] + 1;
    	  if (shb.inEllipsoid(shb.cart_pos) && shb.isDistanced(shb.cart_pos) && (shb.cart_pos[2] >= ball_popping::min_z + shb.BALLR)){
    		  shb.new_pos = true;
    	  }
    	  while(shb.new_pos && !game_exit){
    		  barrett::btsleep(0.02);
    	  }
    	  shb.cart_pos.setZero();
    	  rand_ang.setZero();
          if (game_exit || ps.pressedButton == barrett::SafetyModule::PendantState::ESTOP){ 
	/*const bus::CommunicationsBus& bus1 = pm.getPuck(1)->getBus();
const bus::CommunicationsBus& bus2 = pm.getPuck(2)->getBus();
const bus::CommunicationsBus& bus3 = pm.getPuck(3)->getBus();*/
	//wam.moveTo(wam.getJointPositions());
	//barrett::btsleep(10);
	pm.getPuck(1)->setProperty(pm.getPuck(1)->getBus(), 1, 8, 3);
	pm.getPuck(2)->setProperty(pm.getPuck(2)->getBus(), 2, 8, 3);
	pm.getPuck(3)->setProperty(pm.getPuck(3)->getBus(), 3, 8, 3);
	std::cout<<"game exit"<<std::endl;
//barrett::btsleep(10);
        barrett::systems::disconnect(wam.input);
        wam.idle();
        return 0;
      }
    }
  }
  return 0;
}
