/**
 *	Copyright 2015 Barrett Technology <support@barrett.com>
 *
 *	This file is part of libbarrett.
 *
 *	This version of libbarrett is free software: you can redistribute it
 *	and/or modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This version of libbarrett is distributed in the hope that it will be
 *	useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this version of libbarrett.  If not, see
 *	<http://www.gnu.org/licenses/>.
 *
 *
 *	Barrett Technology Inc.
 *	73 Chapel Street
 *	Newton, MA 02458
 *
 */
/*
 * ex10_haptics.cpp
 *
 *  Created on: Feb 19, 2010
 *  Last Modified on: Dec 11, 2015
 *      Author: dc
 *      Author: Hariharasudan Malaichamee
 */

#include <boost/bind.hpp>
#include <boost/tuple/tuple.hpp>

#include <barrett/exception.h>
#include <barrett/units.h>
#include <barrett/systems.h>
#include <barrett/products/product_manager.h>
#include <barrett/config.h>
#include <proficio/systems/utilities.h>

#include <signal.h>

#include "ex10_haptics.h"

#define BARRETT_SMF_VALIDATE_ARGS
#define NO_CONTROL_PENDANT

#include <proficio/standard_proficio_main.h>

BARRETT_UNITS_FIXED_SIZE_TYPEDEFS;

char* remoteHost = NULL;
double kp = 3e3;
double kd = 3e1;
bool game_exit = false;

bool validate_args(int argc, char** argv) {
	switch (argc) {
	case 4:
		kd = atof(argv[3]);
	case 3:
		kp = atof(argv[2]);
	case 2:
		remoteHost = argv[1];
		break;

	default:
		printf("Usage: %s <otherip> [<kp> [<kd>]]\n", argv[0]);
		return false;
		break;
	}

	printf("Gains: kp = %f; kd = %f\n", kp, kd);
	return true;
}

cf_type scale(boost::tuple<cf_type, double> t) {
	return t.get<0>() * t.get<1>();
}

void exit_pgm(int signum)
{
  game_exit = true;
  return;
}

template <size_t DOF>
int proficio_main(int argc, char** argv, barrett::ProductManager& pm,
             barrett::systems::Wam<DOF>& wam, config side) {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);
	wam.gravityCompensate();
  	barrett::SafetyModule* sm = pm.getSafetyModule();
  	barrett::SafetyModule::PendantState ps;
  	sm->getPendantState(&ps);
	std::string fname="calibration_data/wam3/";
	if(side == LEFT) { // Left Config
		  fname = fname+"LeftConfig.txt";
	  }
	  else if(side == RIGHT) {
		  fname = fname+"RightConfig.txt";
	}
	UserGravComp<DOF> gc(barrett::EtcPathRelative(fname).c_str());
	gc.setGainZero();
    	// instantiate Systems
	NetworkHaptics<DOF> nh(pm.getExecutionManager(), remoteHost, &gc);
	cp_type center;
	center << 0.4, -0.15, 0.0;
	barrett::systems::HapticBall ball(center, 0.12);
	center << 0.35, 0.2, 0.0;
	barrett::math::Vector<3>::type size;
	size << 0.2, 0.2, 0.2;
	barrett::systems::HapticBox box(center, size);

	barrett::systems::Summer<cf_type> dirSum;
	barrett::systems::Summer<double> depthSum;
	barrett::systems::PIDController<double, double> comp;
	barrett::systems::Constant<double> zero(0.0);
	barrett::systems::TupleGrouper<cf_type, double> tg;
	barrett::systems::Callback<boost::tuple<cf_type, double>, cf_type> mult(scale);
	barrett::systems::ToolForceToJointTorques<DOF> tf2jt;
  	barrett::systems::Summer<jt_type, 3> jtSum("+++");
	jt_type jtLimits(35.0);
	jtSat<DOF> jtsat(jtLimits);
        v_type dampingConstants(20.0);
        dampingConstants[2] = 10.0;
	dampingConstants[0] = 30.0;
  	jv_type velocityLimits(1.4);
  	jvSat<DOF> velsat(dampingConstants,velocityLimits);

  	jv_type jvFiltFreq(20.0);
  	barrett::systems::FirstOrderFilter<jv_type> jvFilter;
  	jvFilter.setLowPass(jvFiltFreq);

	// configure Systems
	comp.setKp(kp);
	comp.setKd(kd);

	barrett::systems::modXYZ<cp_type> modcp;
	modcp.negX();
	modcp.negY();
	modcp.xOffset(0.85);
  if (side == LEFT) {
    modcp.yOffset(0.27);
  } else if (side == RIGHT) {
    modcp.yOffset(-0.27);
  }
  modcp.zOffset(-0.2);

	barrett::systems::modXYZ<cf_type> modforce;
	modforce.negX();
	modforce.negY();
  	barrett::systems::connect(wam.jpOutput, gc.input);
  	barrett::systems::connect(wam.jvOutput, jvFilter.input);
	barrett::systems::connect(jvFilter.output, velsat.input);

	if(DOF == 3) { 
	barrett::systems::connect(wam.toolPosition.output, modcp.input);
	barrett::systems::connect(modcp.output, nh.input);
	barrett::systems::connect(modcp.output, ball.input);
	barrett::systems::connect(modcp.output, box.input);

	barrett::systems::connect(mult.output, modforce.input);
	barrett::systems::connect(modforce.output, tf2jt.input);
	} else {
	// connect Systems
	barrett::systems::connect(wam.toolPosition.output, nh.input);
	barrett::systems::connect(wam.toolPosition.output, ball.input);
	barrett::systems::connect(wam.toolPosition.output, box.input);

	barrett::systems::connect(mult.output, tf2jt.input);
	}
	barrett::systems::connect(ball.directionOutput, dirSum.getInput(0));
	barrett::systems::connect(ball.depthOutput, depthSum.getInput(0));

	barrett::systems::connect(box.directionOutput, dirSum.getInput(1));
	barrett::systems::connect(box.depthOutput, depthSum.getInput(1));

	barrett::systems::connect(wam.kinematicsBase.kinOutput, tf2jt.kinInput);
	barrett::systems::connect(dirSum.output, tg.getInput<0>());

	barrett::systems::connect(depthSum.output, comp.referenceInput);
	barrett::systems::connect(zero.output, comp.feedbackInput);
	barrett::systems::connect(comp.controlOutput, tg.getInput<1>());

	barrett::systems::connect(tg.output, mult.input);
  	barrett::systems::connect(tf2jt.output, jtSum.getInput(0));
  	barrett::systems::connect(gc.output, jtSum.getInput(1));
	barrett::systems::connect(velsat.output, jtSum.getInput(2));
  	barrett::systems::connect(jtSum.output, jtsat.input);
	// adjust velocity fault limit
	pm.getSafetyModule()->setVelocityLimit(1.5);
	pm.getSafetyModule()->setTorqueLimit(3.0);
	wam.idle();
	barrett::systems::connect(jtsat.output, wam.input);
	while (true) {
		// block until the user Shift-
		#ifndef NO_CONTROL_PENDANT
		pm.getSafetyModule()->waitForMode(barrett::SafetyModule::IDLE);
		#endif
		if(game_exit){
			pm.getPuck(1)->setProperty(pm.getPuck(1)->getBus(), 1, 8, 3);
			pm.getPuck(2)->setProperty(pm.getPuck(2)->getBus(), 2, 8, 3);
			pm.getPuck(3)->setProperty(pm.getPuck(3)->getBus(), 3, 8, 3);
			barrett::systems::disconnect(wam.input);
			return 0;
		}
    		barrett::btsleep(0.02);
		// block until the user Shift-activates
		#ifndef NO_CONTROL_PENDANT
		pm.getSafetyModule()->waitForMode(barrett::SafetyModule::ACTIVE);
		#endif
	}

	return 0;
}
