/**
 *	Copyright 2016 Barrett Technology <support@barrett.com>
 *
 *	This file is part of libbarrett.
 *
 *	This version of libbarrett is free software: you can redistribute it
 *	and/or modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This version of libbarrett is distributed in the hope that it will be
 *	useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this version of libbarrett.  If not, see
 *	<http://www.gnu.org/licenses/>.
 *
 *
 *	Barrett Technology Inc.
 *	73 Chapel Street
 *	Newton, MA 02458
 *
 */
/*
 * bimanual_wam.h
 *
 *  Created on: March 19, 2016
 *  Last Modified on: May 21, 2016
 *      Author: ab
 */

#ifndef BIMANUAL_WAM_H_
#define BIMANUAL_WAM_H_

#include <iostream>

#include <boost/thread.hpp>
#include <boost/function.hpp>

#include <barrett/config.h>
#include <barrett/exception.h>
#include <barrett/log.h>
#include <barrett/units.h>
#include <barrett/systems.h>
#include <barrett/products/product_manager.h>

#include <proficio/tools/bypass_safety_module.h>

#ifdef NO_CONTROL_PENDANT
	#define BARRETT_SMF_CONFIGURE_PM
	#define BARRETT_SMF_DONT_PROMPT_ON_ZEROING
	#define BARRETT_SMF_DONT_WAIT_FOR_SHIFT_ACTIVATE
#endif

#ifndef BARRETT_SMF_DONT_WAIT_FOR_SHIFT_ACTIVATE
  #define BARRETT_SMF_WAIT_FOR_SHIFT_ACTIVATE true
#else
  #define BARRETT_SMF_WAIT_FOR_SHIFT_ACTIVATE false
#endif

#ifndef BARRETT_SMF_DONT_PROMPT_ON_ZEROING
  #define BARRETT_SMF_PROMPT_ON_ZEROING true
#else
  #define BARRETT_SMF_PROMPT_ON_ZEROING false
#endif


namespace barrett {

template <size_t DOF>
class BimanualWam {
  BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);

 public:
  BimanualWam() : bpm1(NULL), bpm2(NULL), ready1(false), ready2(false),
                  done1(false), done2(false), quitting(false) {}
  ~BimanualWam();
  bool init(std::string, std::string);

  ProductManager* getPm(int num) {
    switch (num) {
      case 1:
        return bpm1;
      case 2:
        return bpm2;
      default:
        return NULL;
    }
  }

  systems::Wam<DOF>* getWam(int num) {
    switch (num) {
      case 1:
        return bpm1->getWam3();
      case 2:
        return bpm2->getWam3();
      default:
        return NULL;
    }
  }

  systems::ExposedOutput<jp_type>* getJpSys(int num) {
    switch (num) {
      case 1:
        return &jp1;
      case 2:
        return &jp2;
      default:
        return NULL;
    }
  }

  systems::ExposedOutput<cp_type>* getCpSys(int num) {
    switch (num) {
      case 1:
        return &cp1;
      case 2:
        return &cp2;
      default:
        return NULL;
    }
  }

  systems::ExposedOutput<jv_type>* getJvSys(int num) {
    switch (num) {
      case 1:
        return &jv1;
      case 2:
        return &jv2;
      default:
        return NULL;
    }
  }

  bool isReady(int num) {
    switch (num) {
      case 1:
        return ready1;
      case 2:
        return ready2;
      default:
        return false;
    }
  }

  bool isDone(int num) {
    switch (num) {
      case 1:
        return done1;
      case 2:
        return done2;
      default:
        return false;
    }
  }

  void quit() {
    quitting = true;
    return;
  }

  bool isQuitting() {
    return quitting;
  }

 private:

  void setReady(int num, bool ready) {
    switch (num) {
      case 1:
        ready1 = ready;
        break;
      case 2:
        ready2 = ready;
        break;
      default:
        break;
    }
  }

  void setDone(int num, bool done) {
    switch (num) {
      case 1:
        done1 = done;
        break;
      case 2:
        done2 = done;
        break;
      default:
        break;
    }
  }

  void wamThread(ProductManager*& bpm,
                 systems::Wam<DOF>& wam,
                 BimanualWam<DOF>* const bimanual,
                 int num);

  boost::thread startWam(ProductManager*& bpm,
                         BimanualWam<DOF>* const bimanual,
                         int num,
                         boost::function<void (ProductManager*&,
                                               systems::Wam<DOF>&,
                                               BimanualWam<DOF>* const,
                                               int)> wt);

#ifdef BARRETT_SMF_CONFIGURE_PM
  bool configure_pm(ProductManager*& bpm);
#endif

  ProductManager *bpm1, *bpm2;
  systems::ExposedOutput<jp_type> jp1, jp2;
  systems::ExposedOutput<cp_type> cp1, cp2;
  systems::ExposedOutput<jv_type> jv1, jv2;
  bool ready1, ready2, done1, done2, quitting;
  boost::thread wt1, wt2;
};

template <size_t DOF>
BimanualWam<DOF>::~BimanualWam() {
  if (bpm1 != NULL) {
    delete bpm1;
  }
  if (bpm2 != NULL) {
    delete bpm2;
  }
}

template <size_t DOF>
bool BimanualWam<DOF>::init(std::string conf1, std::string conf2) {
  installExceptionHandler();

  // Define where the configuration files are located.
  bpm1 = new ProductManager(barrett::EtcPathRelative(conf1).c_str());
  bpm2 = new ProductManager(barrett::EtcPathRelative(conf2).c_str());

  // Start up each individual arm
  printf("Starting robot thread 1...\n");
  wt1 = startWam(bpm1, this, 1, boost::bind(&BimanualWam::wamThread, this, _1, _2, _3, _4));
  printf("Starting robot thread 2...\n");
  wt2 = startWam(bpm2, this, 2, boost::bind(&BimanualWam::wamThread, this, _1, _2, _3, _4));
  return true;
}

template <size_t DOF>
boost::thread BimanualWam<DOF>::startWam(ProductManager*& bpm,
                       BimanualWam<DOF>* const bimanual,
                       int num,
                       boost::function<void (ProductManager*&,
                                             systems::Wam<DOF>&,
                                             BimanualWam<DOF>* const,
                                             int)> wt) {
#ifdef BARRETT_SMF_CONFIGURE_PM
  bimanual->configure_pm(bpm);
#endif

#ifndef NO_CONTROL_PENDANT
  bpm->waitForWam(BARRETT_SMF_PROMPT_ON_ZEROING);
#endif

  bpm->wakeAllPucks();

  if (bpm->foundWam3()) {
    return boost::thread(wt,
                         boost::ref(bpm),
                         boost::ref(*bpm->getWam3(BARRETT_SMF_WAIT_FOR_SHIFT_ACTIVATE)),
                         bimanual,
                         num);
  } else {
    std::cout << "ERROR: No Proficio detected." << std::endl;
  }
}

template <size_t DOF>
void BimanualWam<DOF>::wamThread(ProductManager*& bpm,
                                 systems::Wam<DOF>& wam,
                                 BimanualWam<DOF>* const bimanual,
                                 int num) {
  BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);

  // TODO(ab): move robot startup and systems connect to startWam function.
  // wamThread should only contain the loop that sets values and checks for
  // quit. make a connectSystems function that checks whether connects are
  // successful.
  SafetyModule* sm = bpm->getSafetyModule();

#ifdef NO_CONTROL_PENDANT
  std::cout << "starting wam thread..." << std::flush;
  for (int i=0; i<3; i++) {
    if(sm->getMode() != SafetyModule::ACTIVE){
      std::cout << " " << i << std::flush;
      systems::ExposedOutput<jt_type> zeroTorque;
      jt_type zero(0.0);
      zeroTorque.setValue(zero);
      systems::connect(zeroTorque.output,wam.input);
      usleep(50000);
      bpm->getSafetyModule()->setMode(SafetyModule::ACTIVE);
      if (i >= 2) {
        bpm->getSafetyModule()->waitForMode(SafetyModule::ACTIVE, false);
      } else {
        usleep(200000);
      }
      systems::disconnect(wam.input);
    }
  }
  std::cout << std::endl;
  wam.getLowLevelWam().getPuckGroup().setProperty(Puck::MODE, MotorPuck::MODE_TORQUE);
#endif

  // Get pointers to ExposedOutput systems for this robot
  systems::ExposedOutput<jp_type> *jp = bimanual->getJpSys(num);
  systems::ExposedOutput<cp_type> *cp = bimanual->getCpSys(num);
  systems::ExposedOutput<jv_type> *jv = bimanual->getJvSys(num);

  // Set initial values of ExposedOutput systems
  jp->setValue(wam.getJointPositions());
  cp->setValue(wam.getToolPosition());
  jv->setValue(wam.getJointVelocities());

  // Associate this robot's ExposedOutput systems with other robot's product
  // manager. This ensures that the user of this class is accessing the
  // appropriate combinations of "this robot" and "other robot" when they start
  // connecting systems in their main function. Robot numbers are 1 and 2, so
  // other robot number is 3-num.
  ProductManager* bpmOther = bimanual->getPm(3-num);
  bpmOther->getExecutionManager()->startManaging(*jp);
  bpmOther->getExecutionManager()->startManaging(*cp);
  bpmOther->getExecutionManager()->startManaging(*jv);

  bimanual->setReady(num, true);

  while (bpm->getSafetyModule()->getMode() == SafetyModule::ACTIVE) {
    jp->setValue(wam.getJointPositions());
    cp->setValue(wam.getToolPosition());
    jv->setValue(wam.getJointVelocities());
    if (bimanual->isQuitting()) {
      sm->setMode(barrett::SafetyModule::IDLE);
      break;
    }
    btsleepRT(0.001);
  }

  sm->waitForMode(barrett::SafetyModule::IDLE, false);
  bimanual->setDone(num, true);
}

// Product Manager configuration necessary to preempt the standard main
// function, allowing for pendantless operation.
#ifdef BARRETT_SMF_CONFIGURE_PM
template <size_t DOF>
bool BimanualWam<DOF>::configure_pm(barrett::ProductManager*& bpm) {
  SafetyModule* sm = bpm->getSafetyModule();
  bypassSafetyModule(sm);
  SafetyModule::PendantState ps;

  sm->getPendantState(&ps);
  if (ps.pressedButton == SafetyModule::PendantState::ESTOP) {
    std::cout << "Please Release the ESTOP to start the wam_node" << std::endl;
  }
  while (ps.pressedButton == SafetyModule::PendantState::ESTOP) {
    sm->getPendantState(&ps);
    btsleep(0.1);
  }
  if (sm->getMode() == SafetyModule::ESTOP || sm->getMode() == SafetyModule::IDLE) {
    sm->setMode(SafetyModule::IDLE);
    sm->waitForMode(SafetyModule::IDLE, false);
  }
  return true;
}
#endif

}  // namespace barrett

#endif  // BIMANUAL_WAM_H_
