/**
 *	Copyright 2016 Barrett Technology <support@barrett.com>
 *
 *	This file is part of libbarrett.
 *
 *	This version of libbarrett is free software: you can redistribute it
 *	and/or modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This version of libbarrett is distributed in the hope that it will be
 *	useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this version of libbarrett.  If not, see
 *	<http://www.gnu.org/licenses/>.
 *
 *
 *	Barrett Technology Inc.
 *	73 Chapel Street
 *	Newton, MA 02458
 *
 */
/*
 * ex10_bimanual.cpp
 *
 *  Created on: Feb 19, 2010
 *  Last Modified on: May 21, 2016
 *      Author: dc
 *      Author: ab
 */

#include <boost/bind.hpp>
#include <boost/tuple/tuple.hpp>

#include <barrett/exception.h>
#include <barrett/units.h>
#include <barrett/systems.h>
#include <barrett/products/product_manager.h>

#define NO_CONTROL_PENDANT
#include "bimanual_wam.h"
#include "ex10_haptics.h"


using namespace barrett;
using systems::connect;
BARRETT_UNITS_FIXED_SIZE_TYPEDEFS;


char* remoteHost = NULL;
double kp = 3e3;
double kd = 3e1;
bool validate_args(int argc, char** argv) {
  switch (argc) {
    case 4:
      kd = atof(argv[3]);
    case 3:
      kp = atof(argv[2]);
    case 2:
      remoteHost = argv[1];
      break;

    default:
      printf("Usage: %s <otherip> [<kp> [<kd>]]\n", argv[0]);
      return false;
      break;
  }

  printf("Gains: kp = %f; kd = %f\n", kp, kd);
  return true;
}


cf_type scale(boost::tuple<cf_type, double> t) {
  return t.get<0>() * t.get<1>();
}

typename units::JointTorques<3>::type saturateJt(const typename units::JointTorques<3>::type& x, const typename units::JointTorques<3>::type& limit) {
  int index;
  double minRatio;

  minRatio = (limit.cwise() / (x.cwise().abs())).minCoeff(&index);
  if (minRatio < 1.0) {
    return minRatio * x;
  } else {
    return x;
  }
}

int main(int argc, char** argv) {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(3);
  barrett::BimanualWam<3> robots;

  if (!validate_args(argc, argv)) {
    return 1;
  }

  robots.init("proficio_left/default.conf.0", "proficio_right/default.conf.1");
  while (!robots.isReady(1) || !robots.isReady(2)) {
    barrett::btsleep(0.001);
  }

  int numL = 1, numR = 2;

  // left arm
  barrett::ProductManager *bpmL = robots.getPm(numL);
  barrett::systems::Wam<3> *wamL = robots.getWam(numL);

  // right arm
  barrett::ProductManager *bpmR = robots.getPm(numR);
  barrett::systems::Wam<3> *wamR = robots.getWam(numR);

  // gravity compensate both robots
  wamL->gravityCompensate();
  wamR->gravityCompensate();

  // systems for communication with the python visualization
  NetworkHaptics nhL(bpmL->getExecutionManager(), remoteHost, 5555, 5556);
  NetworkHaptics nhR(bpmR->getExecutionManager(), remoteHost, 5557, 5558);

  // need separate instances of each haptic object for each robot
  cp_type center;
  center << 0.4, -0.15, 0.0;
  double radius = 0.12;
  systems::HapticBall ballL(center, radius);
  systems::HapticBall ballR(center, radius);

  math::Vector<3>::type size;
  center << 0.35, 0.2, 0.0;
  size << 0.2, 0.2, 0.2;
  systems::HapticBox boxL(center, size);
  systems::HapticBox boxR(center, size);

  // Adjust fault limits and torque saturation limit.
  bpmL->getSafetyModule()->setVelocityLimit(1.5);
  bpmR->getSafetyModule()->setVelocityLimit(1.5);
  bpmL->getSafetyModule()->setTorqueLimit(3.0);
  bpmR->getSafetyModule()->setTorqueLimit(3.0);

  jt_type jtLimits(35.0);

  // systems for left arm
  systems::Constant<double> zeroL(0.0);
  systems::Summer<cf_type> dirSumL;
  systems::Summer<double> depthSumL;
  systems::PIDController<double, double> compL;
  systems::TupleGrouper<cf_type, double> tgL;
  systems::Callback<boost::tuple<cf_type, double>, cf_type> multL(scale);
  systems::ToolForceToJointTorques<3> tf2jtL;
  systems::Callback<jt_type> jtSatL(boost::bind(saturateJt, _1, jtLimits));

  // systems for right arm
  systems::Constant<double> zeroR(0.0);
  systems::Summer<cf_type> dirSumR;
  systems::Summer<double> depthSumR;
  systems::PIDController<double, double> compR;
  systems::TupleGrouper<cf_type, double> tgR;
  systems::Callback<boost::tuple<cf_type, double>, cf_type> multR(scale);
  systems::ToolForceToJointTorques<3> tf2jtR;
  systems::Callback<jt_type> jtSatR(boost::bind(saturateJt, _1, jtLimits));

  // set controller gains
  compL.setKp(kp);
  compL.setKd(kd);
  compR.setKp(kp);
  compR.setKd(kd);

  // ModXYZ systems are used for aligning coordinate frames. x and y axes
  // have their direction reversed in the python visualization, so they are
  // negated here for both position and force. For position, offsets are also
  // applied so the workspace of each robot corresponds roughly with the visual
  // workspace. The y offset is different because the two systems are mirrored.
  // The y position offset is chosen so that the two cursors appear in roughly
  // the same place with the robot endpoints are touching, assuming standard
  // separation distance (1100 mm) between the two robot centers.
  systems::modXYZ<cp_type> modcpL;
  systems::modXYZ<cp_type> modcpR;
  modcpL.negX();
  modcpL.negY();
  modcpR.negX();
  modcpR.negY();
  cp_type offsetL(0.85,0.13,-0.2);
  cp_type offsetR(offsetL[0],-1.0*offsetL[1],offsetL[2]);
  modcpL.xOffset(offsetL[0]);
  modcpL.yOffset(offsetL[1]);
  modcpL.zOffset(offsetL[2]);
  modcpR.xOffset(offsetR[0]);
  modcpR.yOffset(offsetR[1]);
  modcpR.zOffset(offsetR[2]);

  systems::modXYZ<cf_type> modforceL;
  systems::modXYZ<cf_type> modforceR;
  modforceL.negX();
  modforceL.negY();
  modforceR.negX();
  modforceR.negY();

  // connect systems to set up the control loop

  // Robot position gets its coordinate axes transformed, then gets sent to
  // the visualization and the haptic object systems
  connect(wamL->toolPosition.output, modcpL.input);
  connect(modcpL.output, nhL.input);
  connect(modcpL.output, ballL.input);
  connect(modcpL.output, boxL.input);

  connect(wamR->toolPosition.output, modcpR.input);
  connect(modcpR.output, nhR.input);
  connect(modcpR.output, ballR.input);
  connect(modcpR.output, boxR.input);

  // Output of haptic object systems is summed. This method only works when
  // the cursor cannot touch multiple objects at the same time. It works here
  // because direction and depth outputs are equal to zero when the cursor is
  // not penetrating the object.
  connect(ballL.directionOutput, dirSumL.getInput(0));
  connect(ballL.depthOutput, depthSumL.getInput(0));
  connect(boxL.directionOutput, dirSumL.getInput(1));
  connect(boxL.depthOutput, depthSumL.getInput(1));

  connect(ballR.directionOutput, dirSumR.getInput(0));
  connect(ballR.depthOutput, depthSumR.getInput(0));
  connect(boxR.directionOutput, dirSumR.getInput(1));
  connect(boxR.depthOutput, depthSumR.getInput(1));

  // PID controller on the depth, trying to drive it to zero. Note again that
  // the depth is zero when the cursor is not touching an object, so the
  // feedback input is only nonzero when the cursor is penetrating the object.
  connect(depthSumL.output, compL.referenceInput);
  connect(zeroL.output, compL.feedbackInput);

  connect(depthSumR.output, compR.referenceInput);
  connect(zeroR.output, compR.feedbackInput);

  // Multiply the force and direction.The input to the "mult" system is a tuple
  // of a vector and a double. The output is their product.
  connect(dirSumL.output, tgL.getInput<0>());
  connect(compL.controlOutput, tgL.getInput<1>());
  connect(multL.output, modforceL.input);

  connect(dirSumR.output, tgR.getInput<0>());
  connect(compR.controlOutput, tgR.getInput<1>());
  connect(multR.output, modforceR.input);

  // Transform the force vector back into the robot frame (negate the x and y
  // axes with a ModXYZ system defined above).
  connect(tgL.output, multL.input);
  connect(tgR.output, multR.input);

  // Calculate the joint torques necessary to apply the specified endpoint
  // force. This requires the robot kinematics.
  connect(wamL->kinematicsBase.kinOutput, tf2jtL.kinInput);
  connect(modforceL.output, tf2jtL.input);

  connect(wamR->kinematicsBase.kinOutput, tf2jtR.kinInput);
  connect(modforceR.output, tf2jtR.input);

  // Apply saturation limits.
  connect(tf2jtL.output, jtSatL.input);
  connect(tf2jtR.output, jtSatR.input);

  // Activate torques to robots.
  connect(jtSatL.output, wamL->input);
  connect(jtSatR.output, wamR->input);

  // create temporary file for startup script
  std::ofstream startFlag;
  startFlag.open("/tmp/csbstart");
  startFlag.close();

  // Run until either robot disconnects, then deactivate both robots.
  std::cout << "Entering main loop... shift-idle either robot to quit." << std::endl;
  while(!robots.isDone(1) && !robots.isDone(2)) {
    barrett::btsleepRT(0.001);
  }
  systems::disconnect(wamL->input);
  systems::disconnect(wamR->input);

  bpmL->getSafetyModule()->setMode(barrett::SafetyModule::IDLE);
  bpmR->getSafetyModule()->setMode(barrett::SafetyModule::IDLE);
  bpmL->getSafetyModule()->waitForMode(barrett::SafetyModule::IDLE, false);
  bpmR->getSafetyModule()->waitForMode(barrett::SafetyModule::IDLE, false);

  return 0;
}
