/**
 *	Copyright 2015 Barrett Technology <support@barrett.com>
 *
 *	This file is part of libbarrett.
 *
 *	This version of libbarrett is free software: you can redistribute it
 *	and/or modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This version of libbarrett is distributed in the hope that it will be
 *	useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this version of libbarrett.  If not, see
 *	<http://www.gnu.org/licenses/>.
 *
 *
 *	Barrett Technology Inc.
 *	73 Chapel Street
 *	Newton, MA 02458
 *
 */

/**
 *  @file haptic_world.cpp
 *  @author Kyle Maroney
 *  @author Hariharasudan Malaichamee
 *  @date 14/12/2015
 */

//Networking */
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

// Barrett Library */
#include <barrett/math.h>
#include <barrett/exception.h>
#include <barrett/units.h>
#include <barrett/systems.h>
#include <barrett/products/product_manager.h>
#include <barrett/config.h>

#include <proficio/systems/utilities.h>

#include "haptic_world.h"

#include <signal.h>     /* signal, raise, sig_atomic_t */

#define BARRETT_SMF_VALIDATE_ARGS
#define NO_CONTROL_PENDANT

#include <proficio/standard_proficio_main.h>

/* Constants describing out environment in meters */
const double BACKWALLX = 0.3;
const double FRONTWALLX = 0.7;
const double LEFTWALLY = -0.25;
const double RIGHTWALLY = 0.25;
const double TOPWALLZ = 0.5;
const double BOTTOMWALLZ = 0.0;

/* Button */
const double BUTTONRAD = .15;
const double BUTTONSIZE = 0.01;

/* Proxy Parameters */
const double PROXYRAD = 0.03;

/* Magnetic field */
const double MAGFIELDSIZE = 0.01;

/* Ball Parameters */
const double BALLSTARTX = 0.45; //0.25;
const double BALLSTARTY = -0.125; //-0.2;
const double BALLSTARTZ = 0.25;
const double BALLR = 0.075;
const double BALLM = 50.0;
const double GRAVITY_FORCE = 0.00015;
const double AIR_DAMPING = 0.0005;

BARRETT_UNITS_FIXED_SIZE_TYPEDEFS;
BARRETT_UNITS_TYPEDEFS(10);

char* remoteHost = NULL;
cf_type currentForce;
cp_type proxyPosition;
cv_type proxyVelocity;
cp_type ballPosition;
cv_type ballVelocity;
v_type msg_tmp;
bool buttonPressed;
barrett::systems::ExposedOutput<v_type> message;
//double stiffness; // Spring Force (Kp)
//double damping; // Damping Force (Kd)
	
/* Verify that IP address is given */	
bool validate_args(int argc, char** argv) {
	if (argc != 2) {
		printf("Usage: %s <otherip>\n", argv[0]);
		return false;
	}
	remoteHost = argv[1];
	return true;
}

cf_type hapticCalc(boost::tuple<cp_type, cv_type> haptuple) {
	cf_type cfSum;
	cp_type wamPosition = haptuple.get<0>();
	cv_type wamVelocity = haptuple.get<1>();

	proxyPosition = wamPosition;
	proxyVelocity = wamVelocity;
	
	double stiffness = 1000.0; // Spring Force (Kp)
	double damping = 15.0; // Damping Force (Kd)
	

	/* Ball Calculations */
	cp_type proxyBallVec = proxyPosition - ballPosition;

	if (proxyBallVec.norm() < (BALLR + PROXYRAD)) {
		proxyPosition = ballPosition
				+ (BALLR + PROXYRAD) * proxyBallVec / proxyBallVec.norm();
		ballVelocity[0] = ballVelocity[0]
				+ (-(stiffness * (proxyPosition[0] - wamPosition[0])) / BALLM)
						* 0.002; // changes the balls velocity when touched by the proxy accordingly in all three dimensions
		ballVelocity[1] = ballVelocity[1]
				+ (-(stiffness * (proxyPosition[1] - wamPosition[1])) / BALLM)
						* 0.002;
		ballVelocity[2] = ballVelocity[2]
				+ (-(stiffness * (proxyPosition[2] - wamPosition[2])) / BALLM)
						* 0.002 - GRAVITY_FORCE;
		; //500Hz
	} else {
		ballVelocity[2] -= GRAVITY_FORCE; //gravity compensation on the ball
		ballVelocity -= AIR_DAMPING * ballVelocity; //damping due to air friction
	}

	cfSum[0] = stiffness * (proxyPosition[0] - wamPosition[0]);
	cfSum[1] = stiffness * (proxyPosition[1] - wamPosition[1]);
	cfSum[2] = stiffness * (proxyPosition[2] - wamPosition[2]);

	//Check for a collision with the ball and right wall using balls radius
	if (ballPosition[1] > RIGHTWALLY - BALLR) {
		ballVelocity[1] = -ballVelocity[1];
	}

	//Check for a collision with the ball and button / left wall using balls radius
	if (ballPosition[1] < LEFTWALLY + BUTTONSIZE + BALLR
			&& ballPosition[0]
					< BACKWALLX + ((FRONTWALLX - BACKWALLX) / 2) + BUTTONRAD
			&& ballPosition[0]
					> BACKWALLX + ((FRONTWALLX - BACKWALLX) / 2) - BUTTONRAD
			&& ballPosition[2]
					< BOTTOMWALLZ + ((TOPWALLZ - BOTTOMWALLZ) / 2) + BUTTONRAD
			&& ballPosition[2]
					> BOTTOMWALLZ + ((TOPWALLZ - BOTTOMWALLZ) / 2) - BUTTONRAD) {
		ballVelocity[1] = -ballVelocity[1];
	} else if (ballPosition[1] < LEFTWALLY + BALLR)
		ballVelocity[1] = -ballVelocity[1];

	// Check for a collision with the ball and ceiling using balls radius
	if (ballPosition[2] > TOPWALLZ - BALLR) {
		ballVelocity[2] = -ballVelocity[2];
	}

	// Check for a collision with the ball and floor using balls radius
	if (ballPosition[2] < BOTTOMWALLZ + BALLR) {
		ballVelocity[2] = -ballVelocity[2];
	}

	// Check for a collision with the ball and front wall using balls radius
	if (ballPosition[0] > FRONTWALLX - BALLR) {
		ballVelocity[0] = -ballVelocity[0];
	}

	// Check for a collision with the ball and back wall using balls radius
	if (ballPosition[0] < BACKWALLX + BALLR) {
		ballVelocity[0] = -ballVelocity[0];
	}

	ballPosition = ballPosition + 0.001 * ballVelocity;

	// Check for a collision with the back wall - Sticky
	if (wamPosition[0] < BACKWALLX + PROXYRAD) {
		proxyPosition[0] = BACKWALLX + PROXYRAD;
		cfSum[0] += 0.6 * (stiffness * (proxyPosition[0] - wamPosition[0])); //Spring Portion
		cfSum[0] -= damping * wamVelocity[0]; // Damping portion
		cfSum[1] -= damping * wamVelocity[1];
		cfSum[2] -= damping * wamVelocity[2];
	}

	// Check for a collision with the front wall - Nothing Keep person in box.
	if (wamPosition[0] > FRONTWALLX - PROXYRAD) {
		proxyPosition[0] = FRONTWALLX - PROXYRAD;
		cfSum[0] += (stiffness * (proxyPosition[0] - wamPosition[0])); //Spring Portion
		cfSum[0] -= damping * wamVelocity[0]; // Damping portion
	}

	// Check for a collision with the right wall - Magnetic
	if (wamPosition[1] > RIGHTWALLY - PROXYRAD - MAGFIELDSIZE) {
		proxyPosition[1] = RIGHTWALLY - PROXYRAD;
		cfSum[1] += (stiffness * (proxyPosition[1] - wamPosition[1])); //Spring Portion
		cfSum[1] -= damping * wamVelocity[1];
	}

	buttonPressed = false;
	// Check for a collision with the left button / WALL
	if (wamPosition[1] < LEFTWALLY + BUTTONSIZE + PROXYRAD
			&& wamPosition[0]
					< BACKWALLX + ((FRONTWALLX - BACKWALLX) / 2) + BUTTONRAD
			&& wamPosition[0]
					> BACKWALLX + ((FRONTWALLX - BACKWALLX) / 2) - BUTTONRAD
			&& wamPosition[2]
					< BOTTOMWALLZ + ((TOPWALLZ - BOTTOMWALLZ) / 2) + BUTTONRAD
			&& wamPosition[2]
					> BOTTOMWALLZ + ((TOPWALLZ - BOTTOMWALLZ) / 2) - BUTTONRAD) {
		proxyPosition[1] = LEFTWALLY + BUTTONSIZE + PROXYRAD;
		if (wamPosition[1] - proxyPosition[1] < -0.01) {
			proxyPosition[1] = LEFTWALLY + PROXYRAD;
			cfSum[1] += (stiffness * (proxyPosition[1] - wamPosition[1])); //Spring Portion
//			cfSum[1] -= damping * wamVelocity[1]; // Damping portion

			buttonPressed = true;
		} else {
			cfSum[1] += (stiffness * (proxyPosition[1] - wamPosition[1])); //Spring Portion
//			cfSum[1] -= damping * wamVelocity[1]; // Damping portion
		}
	} else if (wamPosition[1] < LEFTWALLY + PROXYRAD) {
		proxyPosition[1] = LEFTWALLY + PROXYRAD;
		cfSum[1] += (stiffness * (proxyPosition[1] - wamPosition[1])); //Spring Portion
//		cfSum[1] -= damping * wamVelocity[1]; // Damping portion
	}

	// Check for a collision with the ceiling - Slippery inverse damping
	if (wamPosition[2] > TOPWALLZ - PROXYRAD) {
		proxyPosition[2] = TOPWALLZ - PROXYRAD;
		cfSum[2] += (stiffness * (proxyPosition[2] - wamPosition[2])); //Spring Portion
		cfSum[2] -= damping * wamVelocity[2]; // Damping portion
		cfSum[0] += 0.5 * damping * wamVelocity[0];
		cfSum[1] += 0.25 * damping * wamVelocity[1];
	}

	// Check for a collision with the floor.
	if (wamPosition[2] < BOTTOMWALLZ + PROXYRAD) {
		proxyPosition[2] = BOTTOMWALLZ + PROXYRAD;
		cfSum[2] += (stiffness * (proxyPosition[2] - wamPosition[2]));
		cfSum[2] -= damping * wamVelocity[2];
		if (sin(500 * wamPosition[1]) >= 0)
			cfSum[1] = stiffness * (proxyPosition[1] - wamPosition[1])
					- damping * wamVelocity[1] * 2;
			/* creates the illusion of stripes on the wall by multiplying 
			the position by sin(.5*position), and turns on the parallel 
			damping when it is above 0 and turns it off when it is less than zero */
	}
	return cfSum;
}

enum CONTACT_STATE {
	PLAYING, TARING, QUIT
} curState = PLAYING, lastState = PLAYING;

// HapticsDemo Class
template<size_t DOF>
class HapticsDemo {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);
protected:
  /* Wam and ProductManager Instances Passed into Class */
	barrett::systems::Wam<DOF>& wam;
	barrett::ProductManager& pm;

  /* Network Parameters */
	struct sockaddr_in si_server;
	int port, sock, i, slen;
	char* buf;
	char* srv_addr;
	
	/*  */
	jt_type jtLimits;
	barrett::systems::TupleGrouper<cp_type, cv_type> tg;
	/*  */
  	barrett::systems::ToolForceToJointTorques<DOF> cf_tf2jt;
	
	/*  */	  
	cf_type cfSum;
	jp_type jpStart;
	cf_type curForces;
	/*  */
  	barrett::systems::Callback<boost::tuple<cp_type, cv_type>, cf_type> haptics;
	jtSat<DOF> jtsat;
	UserGravComp<DOF>* gc;
	barrett::systems::Summer<jt_type, 3> jtSum;
	v_type dampingConstants;
	jv_type velocityLimits;
	jvSat<DOF> velsat;

	barrett::systems::modXYZ<cp_type> invpos;
	barrett::systems::modXYZ<cf_type> invforce;
	barrett::systems::modXYZ<cv_type> invvel;
	barrett::systems::modXYZ<cv_type> invvelSat;
  	
	jv_type jvFiltFreq;
  	cv_type cvFiltFreq;
  	barrett::systems::FirstOrderFilter<jv_type> jvFilter;
  	barrett::systems::FirstOrderFilter<cv_type> cvFilter;
public:
	bool ftOn;
	double sumForces;
  /* Constructor */
	HapticsDemo(barrett::systems::Wam<DOF>& wam_, barrett::ProductManager& pm_, UserGravComp<DOF>* ugc) :
			wam(wam_), pm(pm_),slen(sizeof(si_server)), jtLimits(45.0), haptics(hapticCalc), 
                        jtsat(jtLimits), gc(ugc), jtSum("+++"), dampingConstants(20.0), 
                        velocityLimits(1.4), velsat(dampingConstants, velocityLimits), jvFiltFreq(20.0), cvFiltFreq(20.0), sumForces(0.0) {    
        dampingConstants[2] = 10.0;
	dampingConstants[0] = 30.0;
	velsat.setDampingConst(dampingConstants);
	invpos.negX();
	invpos.negY();
	invpos.xOffset(1);
        invforce.negX();
	invforce.negY();
        invvel.negX();
	invvel.negY();
	invvelSat.negX();
	invvelSat.negY();
	jvFilter.setLowPass(jvFiltFreq);
	cvFilter.setLowPass(cvFiltFreq);
		      }
  /* Destructor */
	~HapticsDemo() {}
  /* Public Methods */
	bool init(config side);
	bool setupNetworking();
	void displayEntryPoint();
	void connectForces();
	void visualizationThread();
};

/* Initialization Method */
template<size_t DOF>
bool HapticsDemo<DOF>::init(config side) {
	wam.gravityCompensate();
	pm.getSafetyModule()->setVelocityLimit(1.5);
	pm.getSafetyModule()->setTorqueLimit(2.5);

        if (DOF == 4) {
        	jpStart << 0.0, 0.448, 0.046, 2.178;
        	wam.moveTo(jpStart);
  	} else if (DOF == 3) {
      if (side == LEFT) {
        jpStart << -0.1, 0, -1.5;
      } else if (side == RIGHT) {
        jpStart << -0.1, 0, 1.5;
      }
      wam.moveTo(jpStart);
  	}
	for (int i = 0; i < 3; i++) {
		proxyPosition[i] = 0.0;
		proxyVelocity[i] = 0.0;
		ballVelocity[i] = 0.0;
	}
	ballPosition[0] = BALLSTARTX;
	ballPosition[1] = BALLSTARTY;
	ballPosition[2] = BALLSTARTZ;
	for (int j = 0; j < 10; j++) {
		if (j < 3)
			msg_tmp[j] = proxyPosition[j];
		else if (j < 6)
			msg_tmp[j] = proxyVelocity[j - 3];
		else if (j < 9)
			msg_tmp[j] = ballPosition[j - 6];
		else
			msg_tmp[j] = (double) buttonPressed;
	}
	buttonPressed = false;
  	wam.idle();
	return true;
}

/*  */
template<size_t DOF>
bool HapticsDemo<DOF>::setupNetworking() {
	buf = new char[1024];
	srv_addr = new char[16];
	memset(srv_addr, 0, 16);
	port = 5555;
	if ((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		printf("Failure creating the socket");
		return false;
	}
	memcpy(srv_addr, remoteHost, strlen(remoteHost));
	memset((char *) &si_server, 0, sizeof(si_server));
	si_server.sin_family = AF_INET;
	si_server.sin_port = htons(port);
	if (inet_aton(srv_addr, &si_server.sin_addr) == 0) {
		printf("inet_aton() failed - EXITING\n");
		return false;
	}
	return true;
}

/*  */
template<size_t DOF>
void HapticsDemo<DOF>::connectForces() {
	BARRETT_UNITS_FIXED_SIZE_TYPEDEFS;
	
	barrett::systems::connect(wam.kinematicsBase.kinOutput, cf_tf2jt.kinInput);
	barrett::systems::connect(wam.jpOutput, gc->input);
  	barrett::systems::connect(wam.jvOutput, jvFilter.input);
  	barrett::systems::connect(jvFilter.output, velsat.input); 	
  	//barrett::systems::connect(velsat.output, invvelSat.input); 	
        barrett::systems::connect(wam.toolPosition.output, invpos.input);
	barrett::systems::connect(invpos.output, tg.getInput<0>());
	barrett::systems::connect(haptics.output, invforce.input);
	barrett::systems::connect(invforce.output, cf_tf2jt.input);
        barrett::systems::connect(wam.toolVelocity.output, cvFilter.input);
        barrett::systems::connect(cvFilter.output, invvel.input);
        barrett::systems::connect(invvel.output, tg.getInput<1>());
	barrett::systems::connect(tg.output, haptics.input);
	barrett::systems::connect(cf_tf2jt.output, jtSum.getInput(0));
  	barrett::systems::connect(gc->output, jtSum.getInput(1));
	barrett::systems::connect(velsat.output, jtSum.getInput(2));
	//barrett::systems::connect(invvelSat.output, jtSum.getInput(2));
  	barrett::systems::connect(jtSum.output, jtsat.input);
	barrett::systems::connect(jtsat.output, wam.input);
}

/*  */
template<size_t DOF>
void HapticsDemo<DOF>::visualizationThread() {
	while (curState == PLAYING) {
		sumForces = 0.0;
		for (int i = 0; i < 3; i++)
			sumForces += fabs(curForces[i]);
		for (int j = 0; j < 10; j++) {
			if (j < 3)
				msg_tmp[j] = proxyPosition[j];
			else if (j < 6)
				msg_tmp[j] = proxyVelocity[j - 3];
			else if (j < 9)
				msg_tmp[j] = ballPosition[j - 6];
			else
				msg_tmp[j] = (double) buttonPressed;
		}
		message.setValue(msg_tmp);
		barrett::btsleep(0.01);
	}
}

/*  */
template<size_t DOF>
int proficio_main(int argc, char** argv, barrett::ProductManager& pm,
		barrett::systems::Wam<DOF>& wam, config side) {
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);
	std::string fname="calibration_data/wam3/";
	if(side == LEFT) { // Left Config
		fname = fname+"LeftConfig.txt";
	} else if(side == RIGHT) {
		fname = fname+"RightConfig.txt";
	}
	UserGravComp<DOF> gc(barrett::EtcPathRelative(fname).c_str());
	gc.setGainZero();
	HapticsDemo<DOF> haptics_demo(wam, pm, &gc);
	if (!haptics_demo.setupNetworking()) {
		return 1;
	}
	if (!haptics_demo.init(side))
		return 1;
	// instantiate Systems
	NetworkHaptics<DOF> nh(pm.getExecutionManager(), remoteHost, &gc);
	message.setValue(msg_tmp);
	barrett::systems::forceConnect(message.output, nh.input);
	haptics_demo.connectForces();
	boost::thread visualizeThread(&HapticsDemo<DOF>::visualizationThread,
			&haptics_demo);
	bool running = true;
	haptics_demo.ftOn = false;
	while (running) {
		switch (curState) {
		case QUIT:
			pm.getPuck(1)->setProperty(pm.getPuck(1)->getBus(), 1, 8, 3);
			pm.getPuck(2)->setProperty(pm.getPuck(2)->getBus(), 2, 8, 3);
			pm.getPuck(3)->setProperty(pm.getPuck(3)->getBus(), 3, 8, 3);
			barrett::systems::disconnect(wam.input);
			running = false;
			break;
		case PLAYING:
			lastState = PLAYING;
			barrett::btsleep(0.1);
			break;
		default:
			break;
		}
	}
	wam.moveHome();
	printf("\n\n");
	return 0;
}
