/**
 *	Copyright 2015 Barrett Technology <support@barrett.com>
 *
 *	This file is part of libbarrett.
 *
 *	This version of libbarrett is free software: you can redistribute it
 *	and/or modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This version of libbarrett is distributed in the hope that it will be
 *	useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this version of libbarrett.  If not, see
 *	<http://www.gnu.org/licenses/>.
 *
 *
 *	Barrett Technology Inc.
 *	73 Chapel Street
 *	Newton, MA 02458
 *
 */

/**
 *  @file haptic_world.cpp
 *  @author Kyle Maroney
 *  @author Hariharasudan Malaichamee
 *  @date 14/12/2015
 */
#ifndef HAPTIC_WORLD_H_
#define HAPTIC_WORLD_H_


#include <stdexcept>

#include <syslog.h>
#include <unistd.h> /* for close() */
#include <sys/socket.h> /* For sockets */
#include <fcntl.h>      /* To change socket to nonblocking mode */
#include <arpa/inet.h>  /* For inet_pton() */

#include <barrett/detail/ca_macro.h>
#include <barrett/units.h>
#include <barrett/systems/abstract/single_io.h>
#include <barrett/thread/disable_secondary_mode_warning.h>

template <size_t DOF>
class NetworkHaptics : public barrett::systems::SingleIO<barrett::math::Vector<10>::type, boost::tuple< double, barrett::math::Vector<10>::type> > {
        BARRETT_UNITS_TYPEDEFS(10);

public:
        static const int SIZE_OF_MSG = 10 * sizeof(double);
	static const int SIZE_OF_MSG_RECV = 11 * sizeof(double);

        explicit NetworkHaptics(barrett::systems::ExecutionManager* em, char* remoteHost, UserGravComp<DOF>* gc, int port_src = 5557,
                          int port_dest = 5556, const std::string& sysName = "NetworkHaptics") :
                barrett::systems::SingleIO<v_type, boost::tuple<double, v_type> >(sysName), sock(-1), v(0.0), numMissed(0), cv(0.0), ugc(gc)
        {
                int err;
                long flags;
                int buflen;
                unsigned int buflenlen;
                struct sockaddr_in bind_addr;
                struct sockaddr_in their_addr;
		v_type zero_v(0.0);
		tuple_msg.get<1>() = zero_v;
 		tuple_msg.get<0>() = 10;

                /* Create socket */
                sock = socket(PF_INET, SOCK_DGRAM, 0);
                if (sock == -1)
                {
                        syslog(LOG_ERR,"%s: Could not create socket.",__func__);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }

                /* Set socket to non-blocking, set flag associated with open file */
                flags = fcntl(sock, F_GETFL, 0);
                if (flags < 0)
                {
                        syslog(LOG_ERR,"%s: Could not get socket flags.",__func__);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }
                flags |= O_NONBLOCK;
                err = fcntl(sock, F_SETFL, flags);
                if (err < 0)
                {
                        syslog(LOG_ERR,"%s: Could not set socket flags.",__func__);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }

                /* Maybe set UDP buffer size? */
                buflenlen = sizeof(buflen);
                err = getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&buflen, &buflenlen);
                if (err)
                {
                        syslog(LOG_ERR,"%s: Could not get output buffer size.",__func__);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }
                syslog(LOG_ERR,"%s: Note, output buffer is %d bytes.",__func__,buflen);

                buflenlen = sizeof(buflen);
                buflen = 11 * SIZE_OF_MSG_RECV;
                err = setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&buflen, buflenlen);
                if (err)
                {
                        syslog(LOG_ERR,"%s: Could not set output buffer size.",__func__);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }

                buflenlen = sizeof(buflen);
                err = getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&buflen, &buflenlen);
                if (err)
                {
                        syslog(LOG_ERR,"%s: Could not get output buffer size.",__func__);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }
                syslog(LOG_ERR,"%s: Note, output buffer is %d bytes.",__func__,buflen);

                /* Set up the bind address */
                bind_addr.sin_family = AF_INET;
                bind_addr.sin_port = htons(port_src);
                bind_addr.sin_addr.s_addr = htonl(INADDR_ANY);
                err = bind(sock, (struct sockaddr *)&bind_addr, sizeof(bind_addr));
                if (err == -1)
                {
                        syslog(LOG_ERR,"%s: Could not bind to socket on port %d.",__func__,port_src);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }

                /* Set up the other guy's address */
                their_addr.sin_family = AF_INET;
                their_addr.sin_port = htons(port_dest);
                err = ! inet_pton(AF_INET, remoteHost, &their_addr.sin_addr);
                if (err)
                {
                        syslog(LOG_ERR,"%s: Bad IP argument '%s'.",__func__,remoteHost);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }

                /* Call "connect" to set datagram destination */
                err = connect(sock, (struct sockaddr *)&their_addr, sizeof(struct sockaddr));
                if (err)
                {
                        syslog(LOG_ERR,"%s: Could not set datagram destination.",__func__);
                        throw std::runtime_error("(NetworkHaptics::NetworkHaptics): Ctor failed. Check /var/log/syslog.");
                }


                if (em != NULL) {
                        em->startManaging(*this);
                }
        }

        virtual ~NetworkHaptics() {
                mandatoryCleanUp();
                close(sock);
        }

protected:
        virtual void operate() {
                v = input.getValue();

                {
                        // send() and recv() cause switches to secondary mode. The socket is
                        // non-blocking, so this *probably* won't impact the control-loop
                        // timing that much...
                        barrett::thread::DisableSecondaryModeWarning dsmw;
			//std::cout<<v[0]<<" "<<v[1]<<" "<<v[2]<<std::endl;
                        send(sock, v.data(), SIZE_OF_MSG, 0);
                        recv(sock, &tuple_msg, SIZE_OF_MSG_RECV, 0);
				if (tuple_msg.get<0>() == 0)
					ugc->setGainZero();
				else if (tuple_msg.get<0>() == 1)
					ugc->incrementGain();
				else if (tuple_msg.get<0>() == 2)
					ugc->decrementGain();
           
                }

                outputValue->setData(&tuple_msg);
        }

        int sock;
        v_type v;
        int numMissed;
        v_type cv;
	boost::tuple<double, v_type> tuple_msg;
	UserGravComp<DOF>* ugc;

private:
        DISALLOW_COPY_AND_ASSIGN(NetworkHaptics);
};

#endif /* HAPTIC_WORLD_H_ */

