/* 
 * homecheck.cpp
 *
 * Creator: ab
 * Date: 10/21/15
 *
 * Automatically activates the proficio and checks for home position.
 *
 */

#include <proficio/tools/bypass_safety_module.h>

#include <boost/bind.hpp>
#include <boost/tuple/tuple.hpp>
#include <signal.h>     /* signal, raise, sig_atomic_t */

#include <barrett/exception.h>
#include <barrett/units.h>
#include <barrett/systems.h>
#include <barrett/products/product_manager.h>
#include <proficio/systems/utilities.h>

#include <barrett/math.h>
#include <barrett/systems/wam.h>
#include <barrett/detail/stl_utils.h>
#include <barrett/products/puck.h>
#include <barrett/products/detail/property_list.h>

#define BARRETT_SMF_VALIDATE_ARGS
#define NO_CONTROL_PENDANT

#include <proficio/standard_proficio_main.h>

using namespace barrett;

namespace homecheck {
bool quit = false;
}


//enum robotConfiguration {RIGHT, LEFT};
enum puckVersion {P2, P3};
puckVersion puckType = P3;

bool validate_args(int argc, char** argv) {
	if ((argc != 2) || !((!strcmp(argv[1],"left")) || (!strcmp(argv[1],"right")))) {
		printf("Usage: %s <left/right>\n", argv[0]);
		return false;
	}
	return true;
}

template<size_t DOF>
typename units::JointTorques<DOF>::type saturateJt(const typename units::JointTorques<DOF>::type& x, const typename units::JointTorques<DOF>::type& limit) {
	int index;
	double minRatio;

	minRatio = (limit.cwise() / (x.cwise().abs())).minCoeff(&index);
	if (minRatio < 1.0) {
		return minRatio * x;
	} else {
		return x;
	}
}

void exit_pgm(int signum) {
  homecheck::quit = true;
  std::cout << "quit signal received" << std::endl;
  return;
}

template<size_t DOF>
int proficio_main(int argc, char** argv, barrett::ProductManager& pm,
               barrett::systems::Wam<DOF>& wam, config side)
{
	BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);

  signal(SIGINT, exit_pgm);
  signal(SIGTERM, exit_pgm);
  signal(SIGKILL, exit_pgm);

  SafetyModule* sm = pm.getSafetyModule();

	// for checking if robot is stationary at startup
	jp_type init = wam.getJointPositions();
	usleep(250000);
	sm->setMode(SafetyModule::ACTIVE);
	usleep(250000);

	// check for active state. if failed, try again. if failed twice, exit with error code.
	if (sm->getMode() != SafetyModule::ACTIVE) {
		sm->setMode(SafetyModule::ACTIVE);
		usleep(500000);
		if (sm->getMode() != SafetyModule::ACTIVE) {
      sm->setMode(SafetyModule::ACTIVE);
      usleep(500000);
      if (sm->getMode() != SafetyModule::ACTIVE) {
        return 2;
      }
		}
	}
	sm->setTorqueLimit(3.0);

	// get left or right configuration from command line input
	config arm;
	if (!strcmp(argv[1],"left")) {
		arm = LEFT;
	} else if (!strcmp(argv[1],"right")) {
		arm = RIGHT;
	} else {
		//std::cout << "Invalid configuration. Quitting." << std::endl;
		sm->setMode(SafetyModule::IDLE);
		sm->waitForMode(SafetyModule::IDLE, false);
		return 3;
	}

	// set up torque limits for testing
	const jt_type zeroT(0.0);
	jt_type T = zeroT;
	jt_type Tmax;
	if (puckType == P2) {
		Tmax << 11.0, 9.0, 9.0;
	} else if (puckType == P3) {
		Tmax << 20.0, 22.0, 12.0;
	} else {
		sm->setMode(SafetyModule::IDLE);
		sm->waitForMode(SafetyModule::IDLE, false);
		return 5;
	}

	// velocity commands for testing
	jv_type cmdVel;

	if (arm == LEFT) {
		// home position is -1.01, 0.96, -2.59.
		// for J3, look for opposite home, which is at -0.14.
		// command velocities toward joint stops.
		if (puckType == P2) {
			cmdVel << -0.3, 0.3, 0.3;
		} else {
			cmdVel << -0.4, 0.4, 0.4;
		}
	} else if (arm == RIGHT) {
		// home position is -1.01, -0.96, 2.59. 
		// for J3, look for opposite home, which is at 0.14.
		// command velocities toward joint stops.
		if (puckType == P2) {
			cmdVel << -0.3, -0.3, -0.3;
		} else {
			cmdVel << -0.4, -0.4, -0.4;
		}
	} else {
		sm->setMode(SafetyModule::IDLE);
		sm->waitForMode(SafetyModule::IDLE, false);
		return 4;
	}

	// set up velocity controller
	systems::ExposedOutput<jv_type> refVel;
	systems::PIDController<jv_type,jt_type> pidVel;
	jv_type Kp, Ki, zero, KpON, KiON;
	zero << 0, 0, 0;
	Kp = zero;
	Ki = zero;
	if (puckType == P2) {
		KpON << 15, 35, 10;
		KiON << 250, 55, 50;
	} else {
		KpON << 20, 45, 13;
		KiON << 450, 130, 80;
	}
	pidVel.setKp(Kp);
	pidVel.setKd(zero);
	pidVel.setKi(Ki);
	refVel.setValue(zero);
	systems::connect(wam.jvOutput, pidVel.feedbackInput);
	systems::connect(refVel.output, pidVel.referenceInput);

	// set up position controller - to test one joint at a time, hold other joints stationary
	systems::ExposedOutput<jp_type> refPos;
	systems::PIDController<jp_type,jt_type> pidPos;
	jp_type Kpp, Kdp, zerop, KppON, KdpON;
	zerop << 0, 0, 0;
	Kpp = zerop;
	Kdp = zerop;
	if (puckType == P2) {
		KppON << 500, 400, 400;
		KdpON << 2,2,2;
	} else {
		KppON << 2000, 1200, 600;
		KdpON << 5,5,5;
	}
	pidPos.setKp(Kpp);
	pidPos.setKd(Kdp);
	pidPos.setKi(zerop);
	refPos.setValue(wam.getJointPositions());
	systems::connect(wam.jpOutput, pidPos.feedbackInput);
	systems::connect(refPos.output, pidPos.referenceInput);

	// torque summer
	systems::Summer<jt_type,2> jtSum("++");
	systems::connect(pidVel.controlOutput, jtSum.getInput(0));
	systems::connect(pidPos.controlOutput, jtSum.getInput(1));

	// torque saturation
	systems::Callback<jt_type> jtSat(boost::bind(saturateJt<DOF>, _1, Tmax));
	systems::connect(jtSum.output, jtSat.input);

	if ((sm->getMode() == 1) || ((wam.getJointPositions()-init).norm()) > 0.1) {
		//std::cout << "Fault detected. Quitting." << std::endl;
		sm->setMode(SafetyModule::IDLE);
		sm->waitForMode(SafetyModule::IDLE, false);
		return 6;
	}
	wam.idle();
	// activate torques to robot (should be zero right now)
	systems::connect(jtSat.output, wam.input);

	// test joints one at a time, in the order 2,3,1.
	// this is necessary because the J1 torque needed to lift the arm in a
	// fully extended position is higher than the torque that drives J1 into
	// joint stop so far that plastic hits metal. but when J2 is home, the
	// required J1 torque is acceptable. J3 should be tested before J2 to
	// reduce the likelihood that the cradle will hit the link when the cradle
	// is not locked.

	int nSteps = 100;
	jv_type increment = cmdVel/((double)nSteps);
	jv_type velocity;

	int order[3] = {2,3,1};
	for (size_t k=0; k<DOF; k++) {
		int j = order[k] - 1;
		velocity = zero;
		refVel.setValue(velocity);

		// turn on/off appropriate gains
		for (int i=0; i<3; i++) {
			if (i==j) {
				// turn on velocity control, turn off position control
				Kpp[i] = 0;
				Kdp[i] = 0;
				Kp[i] = KpON[i];
				Ki[i] = KiON[i];
			} else {
				// turn off velocity control, turn on position control
				Kp[i] = 0;
				Ki[i] = 0;
				Kpp[i] = KppON[i];
				Kdp[i] = KdpON[i];
			}
		}
		pidVel.setKp(Kp);
		pidVel.setKi(Ki);
		pidVel.setKd(zero);
		pidPos.setKp(Kpp);
		pidPos.setKd(Kdp);
		pidPos.setKi(zerop);

		// turn on hold position
		refPos.setValue(wam.getJointPositions());

		// ramp up commanded velocity
		for (int step=0; step<nSteps; step++) {
			velocity[j] += increment[j];
			refVel.setValue(velocity);
			if (step%10==0) {
				if (sm->getMode() == 1) {
					//std::cout << "\nFault detected." << std::endl;
					homecheck::quit = true;
					break;
				}
			}
			btsleep(0.001);
		}

		if (homecheck::quit) {
			break;
		}

		// try to move at constant velocity for some amount of time
		for (int i=0; i<50; i++) {
			if (sm->getMode() == 1) {
				//std::cout << "\nFault detected." << std::endl;
				homecheck::quit = true;
				break;
			}
			btsleep(0.01);
		}

		if (homecheck::quit) {
			break;
		}

		// check joint velocity
		velocity = wam.getJointVelocities();
		double delta;
		if (puckType == P2) {
			delta = 0.1;
		} else {
			delta = 0.2;
		}
		if (fabs(velocity[j]) >= delta) {
			homecheck::quit = true;
		} else {
		}

		// ramp commanded velocity back down for a graceful exit
		for (int step=0; step<nSteps; step++) {
			velocity[j] -= increment[j];
			refVel.setValue(velocity);
			if (step%10==0) {
				if (sm->getMode() == 1) {
					homecheck::quit = true;
					break;
				}
			}
			btsleep(0.001);
		}

		refVel.setValue(zero);
		btsleep(0.2);

		if (homecheck::quit) {
			break;
		}
	}

	// If robot is in correct position, (J1, J2 homed, J3 opposite home),
	// then slowly move J3 to home position. If J3 doesn't make it to home,
	// that means the cradle is in the way, so return a failure. Put limits
	// on the torque here to avoid breaking the cradle if it is in the way.

	if (!homecheck::quit) {
		// turn on position control for all joints
		pidVel.setKp(zero);
		pidVel.setKi(zero);
		refVel.setValue(zero);
		jp_type start = wam.getJointPositions();
		refPos.setValue(start);
		pidPos.setKp(KppON);
		pidPos.setKd(KdpON);
		jt_type Tmax2;
		Tmax2 << 8,8,4;
		systems::Callback<jt_type> jtSat2(boost::bind(saturateJt<DOF>, _1, Tmax2));
		systems::connect(pidPos.controlOutput, jtSat2.input);
		systems::disconnect(wam.input);
		systems::connect(jtSat2.output, wam.input);

		double j3range = 2.59-0.14; // expected distance to travel in J3 based on published joint limits
		double j3vel = -1.0*cmdVel[2];
		double dt = 0.01;
		jp_type position = start;

		bool home = false;
		while (!home) {
			if (sm->getMode() == 1) {
				homecheck::quit = true;
				break;
			}
			position[2] += j3vel*dt;
			refPos.setValue(position);
			btsleep(dt);
			// check distance moved - look for commanded position a little bit into joint stop
			if (j3range - fabs(position[2] - start[2]) <= -0.1) {
				home = true;
			}
		}
		btsleep(0.2); // wait for position to settle
		// check position error
		if (fabs(position[2] - (wam.getJointPositions())[2]) > 0.05) {
			homecheck::quit = true;
		}
	}

	// quit
	systems::disconnect(wam.input);
	btsleep(0.5);
	if (!homecheck::quit) 
	{
		pm.getSafetyModule()->getPuck()->setProperty(Puck::ZERO, 0);
		sm->setMode(SafetyModule::IDLE);
		sm->waitForMode(SafetyModule::IDLE, false);
		wam.getLowLevelWam().definePosition(wam.getHomePosition());
		sm->setMode(SafetyModule::ACTIVE);
		sm->waitForMode(SafetyModule::ACTIVE, false);
		wam.gravityCompensate(true);
		pm.getSafetyModule()->setMode(SafetyModule::ESTOP);
	    	wam.getLowLevelWam().getPuckGroup().setProperty(Puck::MODE, MotorPuck::MODE_IDLE);

		// make sure it set home successfully
		/*if ((wam.getJointPositions() - wam.getHomePosition()).norm() > 0.2) 
		{
			sm->setWamZeroed(false);
			pm.getSafetyModule()->setMode(SafetyModule::ESTOP);
	    		wam.getLowLevelWam().getPuckGroup().setProperty(Puck::MODE, MotorPuck::MODE_IDLE);
			return 0;
		} */

			return 1;
	} else 
	{
		sm->setMode(SafetyModule::IDLE);
		sm->waitForMode(SafetyModule::IDLE, false);
			pm.getSafetyModule()->setMode(SafetyModule::ESTOP);
	    		wam.getLowLevelWam().getPuckGroup().setProperty(Puck::MODE, MotorPuck::MODE_IDLE);
		return 2;
	}
	if(pm.getSafetyModule()->getMode() == SafetyModule::ESTOP || pm.getSafetyModule()->getMode() == SafetyModule::IDLE){
		return 3;
	}
}
