/*
 * activate_robot.cpp
 *
 *  Created on: Oct 7, 2015
 *      Author: Hariharasudan Malaichamee
 */

#include <barrett/math.h>
#include <barrett/units.h>
#include <barrett/systems.h>
#include <barrett/products/product_manager.h>
#include <barrett/products/safety_module.h>
#include <barrett/systems/wam.h>
#include <barrett/detail/stl_utils.h>
#include <proficio/systems/utilities.h>

#include <signal.h>     /* signal, raise, sig_atomic_t */

#define BARRETT_SMF_VALIDATE_ARGS
#define NO_CONTROL_PENDANT

#include <proficio/standard_proficio_main.h> 

using namespace barrett;
std::string requestedState = "";
bool exit_pgm = false;

bool validate_args(int argc, char** argv) {
  switch (argc) {
    case 2:
      requestedState = argv[1];
      if (requestedState != "Activate" && requestedState != "Sleep"){
        std::cerr<<"Invalid option";
	return false;
      }
      break;

    default:
      printf("Usage: %s <Activate/Sleep> \n", argv[0]);
      return false;
      break;
  }
  return true;
}

void exit_pgm_cb(int signum)
{
  exit_pgm = true;
  return;
}

class HorizontalFloor : public systems::System, public systems::SingleOutput<units::CartesianForce::type> {
  BARRETT_UNITS_FIXED_SIZE_TYPEDEFS;
 protected:
  cp_type objectCenter, proxyCenter, toolCenter;
  double magField, kp, kd, toolradius;
  cf_type cfSum;
  cv_type wamVelocity;
 public:
  Input<cp_type> cPosition;
  Input<cv_type> cVelocity;
  HorizontalFloor(const std::string& sysName = "MagneticObject") :
    System(sysName), SingleOutput<cf_type>(this), cPosition(this), cVelocity(this), kp(0), kd(0.5), toolradius(0.02){
    objectCenter.setZero();
    proxyCenter.setZero();
    cfSum.setZero();
  }
  void setCenter(cp_type &center){
    objectCenter = center;
    return;
  }
  void setToolRadius(double &radius){
    toolradius = radius;
    return;
  }
  void setkp(double &gain){
    kp = gain;
    return;
  }
 protected:
  void operate(){
    toolCenter = cPosition.getValue();
    wamVelocity = cVelocity.getValue();
    cfSum.setZero();
    if (toolCenter[2] <= (objectCenter[2] + toolradius)) {
      proxyCenter[2] = objectCenter[2] + toolradius;
       cfSum[2] += kp * (proxyCenter[2] - toolCenter[2]);
       cfSum[2] -= kd * (wamVelocity[2]);
     }
    outputValue->setData(&cfSum);
  }
};

//wam_main Function
template<size_t DOF>
int proficio_main(int argc, char** argv, barrett::ProductManager& pm,
               barrett::systems::Wam<DOF>& wam, config side)
{
    BARRETT_UNITS_TEMPLATE_TYPEDEFS(DOF);
    wam.gravityCompensate(true);
    // systems::ExposedOutput<jt_type> zeroTorque;
    //jt_type zero(0.0);
    //zeroTorque.setValue(zero);
    //systems::connect(zeroTorque.output,wam.input);

    v_type dampingConstants(0.0);
    dampingConstants << 32.0,22.0,17.0;
    pm.getSafetyModule()->setVelocityLimit(1.5);
    pm.getSafetyModule()->setTorqueLimit(3.0);
    jv_type velocityLimits(1.3);
    jvSat<DOF> velsat(dampingConstants,velocityLimits,2);

    jv_type freq_jvsat;
    freq_jvsat << 20, 20, 20;
    systems::FirstOrderFilter<jv_type> filter_jvsat;
    filter_jvsat.setLowPass(freq_jvsat);

    systems::connect(wam.jvOutput, filter_jvsat.input);
    systems::connect(filter_jvsat.output, velsat.input);

    signal(SIGINT, exit_pgm_cb);
    signal(SIGTERM, exit_pgm_cb);
    signal(SIGKILL, exit_pgm_cb);

    usleep(50000);
    //pm.getSafetyModule()->setMode(SafetyModule::ACTIVE);
    //pm.getSafetyModule()->waitForMode(SafetyModule::ACTIVE, false);
    //wam.getLowLevelWam().getPuckGroup().setProperty(Puck::MODE, MotorPuck::MODE_TORQUE);

    //systems::disconnect(wam.input);
    //wam.moveTo(wam.getJointPositions());
    btsleep(2);	
    usleep(500000);
    cp_type axis, center, pos;
    cv_type freq;
    double magfield = 0, floorForce = 1000, toolRadius = 0.02, noFloor = 100, max_height = 0.15;
    axis << 1, 0, 0;
    freq << 20, 20, 20;
    jt_type jtLimits(35.0);

    jp_type home;
    bool under = false;
    jp_type jpos = wam.getJointPositions();
    if(jpos[2] < 0) { // Left Config
      home << -1.01, 0.96, -2.84;
    }
    else if(jpos[2] > 0) {
      home << -1.01, -0.96, 2.84;
    }
    else {
      printf("No compatible robot was found ");
      return 0;
    }
    pos = wam.getToolPosition();
    if(pos[2] <= max_height){
      center << 0 , 0, pos[2] - toolRadius;
      under = true;
    }
    else{
      center << 0 , 0, max_height - toolRadius;
    }
    pm.getSafetyModule()->setTorqueLimit(4);

    systems::FirstOrderFilter<cv_type> fof;
    fof.setLowPass(freq);
    HorizontalFloor virtualFloor;
    virtualFloor.setCenter(center);
    virtualFloor.setToolRadius(toolRadius);
    virtualFloor.setkp(noFloor);
    double ang = 0.78;
    systems::ToolForceToJointTorques<DOF> cf_tf2jt;
    systems::Summer<jt_type> jtSum;
    jtSat<DOF> jtsat(jtLimits);

    systems::connect(wam.toolPosition.output, virtualFloor.cPosition);
    systems::connect(wam.toolVelocity.output, fof.input);
    systems::connect(fof.output, virtualFloor.cVelocity);
    systems::connect(virtualFloor.output, cf_tf2jt.input);
    systems::connect(wam.kinematicsBase.kinOutput, cf_tf2jt.kinInput);
    systems::connect(cf_tf2jt.output, jtSum.getInput(0));
    systems::connect(velsat.output, jtSum.getInput(1));
    systems::connect(jtSum.output, jtsat.input);
    usleep(250000);
    pm.getSafetyModule()->setMode(SafetyModule::ACTIVE);
    usleep(250000);
    systems::connect(jtsat.output, wam.input);
    wam.idle();
    double zero_gain = 0;
    while(true)
    {
      pos = wam.getToolPosition();
      jpos = wam.getJointPositions();
      if (under && pos[2] > (center[2] + toolRadius)){ // Keep moving the floor up till a height.
        center << center[0], center[1], pos[2] - toolRadius;
        virtualFloor.setCenter(center);
        if(pos[2] > max_height)
          under = false;
      }
	if(exit_pgm){
	pm.getPuck(1)->setProperty(pm.getPuck(1)->getBus(), 1, 8, 3);
	pm.getPuck(2)->setProperty(pm.getPuck(2)->getBus(), 2, 8, 3);
	pm.getPuck(3)->setProperty(pm.getPuck(3)->getBus(), 3, 8, 3);
          systems::disconnect(wam.input);
		wam.idle();
		return 1;
	}
        if(((requestedState == "Sleep") && (fabs(jpos[0]-home[0]) <= 0.05 && fabs(jpos[1]-home[1]) <= 0.05 && fabs(jpos[2]-home[2]) <= 0.05))){
          // {TODO} Send a CAN message asking the Pucks to power off
	    pm.getSafetyModule()->setMode(SafetyModule::ESTOP);
	    wam.getLowLevelWam().getPuckGroup().setProperty(Puck::MODE, MotorPuck::MODE_IDLE);
          return 1;
        }
	if(pm.getSafetyModule()->getMode() == SafetyModule::ESTOP || pm.getSafetyModule()->getMode() == SafetyModule::IDLE){
		return 3;
	}
      usleep(10000);
    }

    return 2;
}
